/*****************************************************************************
*
* Name: debug.h
* Project: SDIO API export driver
* Version: $Revision: 1.1 $
* Date:    $Date: 2007/01/18 09:31:25 $
* Purpose: 
*
*
*****************************************************************************/
/******************************************************************************
 *
 * (C)Copyright 2005 Marvell
 *
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF MARVELL
 * The copyright notice above does not evidence any
 * actual or intended publication of such source code.
 *
 * This Module contains Proprietary Information of Marvell
 * and should be treated as Confidential.
 *
 * The information in this file is provided for the exclusive use of
 * the licensees of Marvell.
 * Such users have the right to use, modify, and incorporate this code
 * into products for purposes authorized by the license agreement
 * provided they include this notice and the associated copyright notice
 * with any such product.
 * The information in this file is provided "AS IS" without warranty.
 *
 ******************************************************************************/
/*****************************************************************************
*
* History:
* $Log: debug.h,v $
* Revision 1.1  2007/01/18 09:31:25  pweber
* Put under CVS control
*
*
*****************************************************************************/

#ifndef _DEBUG_H_
#define _DEBUG_H_

#ifdef WINNT
#include <wdm.h>
#endif

#define DBG_NONE        0x00000000      // No debug - all off.
#define DBG_ERROR       0x00000001      // Debug for ERRORS
#define DBG_WARNING     0x00000002      // Debug for Warnings.
#define DBG_LOAD        0x00000004      // Debug for driver load
#define DBG_HWIF	      0x00000008
#define DBG_IRQ					0x00000080

#define DBG_T1          0x01000000      // Our various time stamps.. up to 6.
#define DBG_T2          0x02000000      //  ||
#define DBG_T3          0x04000000      //  ||
#define DBG_T4          0x08000000      //  ||
#define DBG_T5          0x10000000      //  ||
#define DBG_T6          0x20000000      //  \/
#define DBG_ALL         0xffffffff      // All debug enabled.

#define DBG_TIMESTAMPS ( DBG_T1 |  DBG_T2  | DBG_T3 | DBG_T4 | DBG_T5 | DBG_T6 )

#define DBG_DEFAULT (DBG_NONE \
| DBG_ERROR      \
| DBG_WARNING    \
| DBG_LOAD  \
| DBG_HWIF   \
)

#ifdef DBG
extern ULONG apiDebugLevel;
#ifdef WINNT
extern unsigned char lptRegImage1;
extern unsigned char lptRegImage2;
extern LARGE_INTEGER t1Time, t1LastTime, t1CurTime;
extern LARGE_INTEGER t2Time, t2LastTime, t2CurTime;
extern LARGE_INTEGER t3Time, t3LastTime, t3CurTime;
extern LARGE_INTEGER t4Time, t4LastTime, t4CurTime;
extern LARGE_INTEGER t5Time, t5LastTime, t5CurTime;
extern LARGE_INTEGER t6Time, t6LastTime, t6CurTime;
#endif // WINNT
#endif // #ifdef DBG

#ifdef WINNT
NTHALAPI
    LARGE_INTEGER
KeQueryPerformanceCounter(OUT PLARGE_INTEGER PerformanceFrequency OPTIONAL);

extern LARGE_INTEGER isrTimeDelay;
#endif

#ifdef DBG

#define DBG_CRLF        0x00000000      // add some cr lfs before this message.
#define DBG_RAW         0x00000000      // raw debug output
#define DBG_DONTCARE (DBG_RAW | DBG_CRLF)       // either of these by themselves should not cause debug to print.

#define BUGPRINT(Str) DbgPrint Str
#ifdef WINNT

#define KERN_DEBUG

#define INITDEBUG() \
    KeQuerySystemTime(&t1Time); \
    KeQuerySystemTime(&t2Time); \
    KeQuerySystemTime(&t3Time); \
    KeQuerySystemTime(&t4Time); \
    KeQuerySystemTime(&t5Time); \
    KeQuerySystemTime(&t6Time); \
    KeQuerySystemTime(&t1CurTime); \
    KeQuerySystemTime(&t2CurTime); \
    KeQuerySystemTime(&t3CurTime); \
    KeQuerySystemTime(&t4CurTime); \
    KeQuerySystemTime(&t5CurTime); \
    KeQuerySystemTime(&t6CurTime); \
    KeQuerySystemTime(&t1LastTime); \
    KeQuerySystemTime(&t2LastTime); \
    KeQuerySystemTime(&t3LastTime); \
    KeQuerySystemTime(&t4LastTime); \
    KeQuerySystemTime(&t5LastTime); \
    KeQuerySystemTime(&t6LastTime); \

#define PrintMacro DbgPrint

#define DBGPRINT(lvl, Str)                              \
    {                                                   \
    ULONG __lvl = lvl;                                  \
    if ( apiDebugLevel & (__lvl & ~DBG_DONTCARE))            \
    {                                                   \
        if ( lvl & DBG_CRLF )                           \
        {                                               \
            PrintMacro("\n");                         \
        }                                               \
				KeQuerySystemTime(&t2CurTime);       \
				PrintMacro("TS:%p:%d:%I64d - ",KeGetCurrentThread(),KeGetCurrentIrql(),(t2CurTime.QuadPart - t2LastTime.QuadPart)/10000); \
				t2LastTime.QuadPart = t2CurTime.QuadPart;   \
        if ( __lvl & DBG_T1 )                             \
        {                                               \
            KeQuerySystemTime(&t1CurTime);       \
            PrintMacro("T1:%I64d - ",t1CurTime.QuadPart - t1LastTime.QuadPart); \
            t1LastTime.QuadPart = t1CurTime.QuadPart;   \
        }                                               \
        else if ( __lvl & DBG_T3 )                        \
        {                                               \
            KeQuerySystemTime(&t3CurTime);       \
            PrintMacro("T3:%I64d - ",t3CurTime.QuadPart - t3LastTime.QuadPart); \
            t3LastTime.QuadPart = t3CurTime.QuadPart;   \
        }                                               \
        else if ( __lvl & DBG_T4 )                        \
        {                                               \
            KeQuerySystemTime(&t4CurTime);       \
            PrintMacro("T4:%I64d - ",t4CurTime.QuadPart - t4LastTime.QuadPart); \
            t4LastTime.QuadPart = t4CurTime.QuadPart;   \
        }                                               \
        else if ( __lvl & DBG_T5 )                        \
        {                                               \
            KeQuerySystemTime(&t5CurTime);       \
            PrintMacro("T5:%I64d - ",t5CurTime.QuadPart - t5LastTime.QuadPart); \
            t5LastTime.QuadPart = t5CurTime.QuadPart;   \
        }                                               \
        else if ( __lvl & DBG_T6)                         \
        {                                               \
            KeQuerySystemTime(&t6CurTime);       \
            PrintMacro("T6:%I64d - ",t6CurTime.QuadPart - t6LastTime.QuadPart); \
            t6LastTime.QuadPart = t6CurTime.QuadPart;   \
        }                                               \
        PrintMacro Str;                                   \
    }}
#else // #ifdef WINNT
#define PrintMacro printk

#define INITDEBUG()

#define DBGPRINT(lvl, Str)                             \
    {                                                   \
    ULONG __lvl = lvl;                                  \
    if ( apiDebugLevel & (__lvl & ~DBG_DONTCARE))            \
    {                                                   \
        PrintMacro Str ;                                   \
    }}

#endif // #ifdef WINNT
#else
#define DBGPRINT(lvl,Str)
#define INITDEBUG()
#endif
#endif
