/** \file sdio_if.h
 *
 * Name:	sdio_if.h
 * Project:	Wireless LAN, Interface to SDIO bus driver
 * Version:	$Revision: 1.1 $
 * Date:	$Date: 2007/01/18 09:31:25 $
 * Purpose:	SDIO interface 2.0 to bus driver definitions
 *
 ******************************************************************************/

/******************************************************************************
 *
 *	(C)Copyright 2005 Marvell
 *
 *	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SYSKONNECT
 *	The copyright notice above does not evidence any
 *	actual or intended publication of such source code.
 *
 *	This Module contains Proprietary Information of SysKonnect
 *	and should be treated as Confidential.
 *
 *	The information in this file is provided for the exclusive use of
 *	the licensees of SysKonnect.
 *	Such users have the right to use, modify, and incorporate this code
 *	into products for purposes authorized by the license agreement
 *	provided they include this notice and the associated copyright notice
 *	with any such product.
 *	The information in this file is provided "AS IS" without warranty.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * History:
 *
 *	$Log: sdio_if.h,v $
 *	Revision 1.1  2007/01/18 09:31:25  pweber
 *	Put under CVS control
 *	
 *
 ******************************************************************************/
#ifndef __SDIO_IF_H
#define __SDIO_IF_H

#ifndef WINNT
#include        <linux/version.h>
#endif

                                                                                                                                                                        /*--- mmoser 3/20/2006 ---*/
#include	"sktypes.h"
#include	"sdio.h"

#ifdef WINNT
#pragma once
#endif

#ifdef WINNT
#ifdef SDIO_API_INTERNAL
#define SDIO_API_IMPORT
#else
#define SDIO_API_IMPORT DECLSPEC_IMPORT
#endif // #ifdef SDIO_API_INTERNAL
#else
#define SDIO_API_IMPORT
#define EXTERN_C	extern
#endif

#ifdef USE_IF_SPINLOCK
#ifndef WINNT
#define LOCK_IF() unsigned long flags; \
	spin_lock_irqsave(&pSdio->IFSpinLock, flags);

#define UNLOCK_IF() spin_unlock_irqrestore(&pSdio->fwSpinLock, flags);
#else
#define LOCK_IF()
#define UNLOCK_IF()
#endif
#else
#define LOCK_IF()
#define UNLOCK_IF()
#endif

#ifdef USE_FW_SPINLOCK
#ifndef WINNT
#define LOCK_FW() unsigned long flags; \
	spin_lock_irqsave(&pSdio->IFSpinLock, flags);

#define UNLOCK_FW() spin_unlock_irqrestore(&pSdio->fwSpinLock, flags);
#else
#define LOCK_FW()
#define UNLOCK_FW()
#endif
#else
#define LOCK_FW()
#define UNLOCK_FW()
#endif

#ifndef WINNT
/*
// Call back functions for WLAN module
typedef void *(* add_card_notifier)(void *dev_id);
typedef int (* remove_card_notifier)(void *dev_id);
*/

// Call back functions for busdriver module
typedef void (*sdio_add_card_notifier) (void *);
typedef void (*sdio_remove_card_notifier) (void);

void mrvlsdio_register_callbacks(sdio_add_card_notifier add,
                                 sdio_remove_card_notifier _remove);
int mrvlsdio_get_interface(PGUID guid, ULONG size, ULONG version,
                           PVOID interface);

#endif

#define MILLI_SECOND 10000

#ifdef WINNT
#define _SLEEP_MS( pDev, millisecs ){\
LARGE_INTEGER duetime;\
if (KeGetCurrentIrql() < 2){\
duetime.QuadPart = -((LONGLONG)MILLI_SECOND * (LONGLONG)(millisecs));\
KeSetTimer ( &(pDev)->SleepTimer,duetime, NULL );\
KeWaitForSingleObject( &(pDev)->SleepTimer,Executive,KernelMode,FALSE,NULL );\
}else{\
KeStallExecutionProcessor(millisecs*1000);}\
}
#else
#define _SLEEP_MS( pDev, millisecs ) \
mdelay(millisecs)
#endif

/** \enum SDIO_API_STATUS
 Status/ error codes returned by the abstracted interfaces 
*/
typedef enum
{
    SDIO_API_STATUS_SUCCESS = 0,
    SDIO_API_STATUS_DEVICE_BUSY,        //!< Device is busy... must wait.
    SDIO_API_STATUS_NO_RESP,    //!< No response is waiting from the device.
    SDIO_API_STATUS_NO_DATA,    //!< No data available at the device.
    SDIO_API_STATUS_RESOURCES,  //!< Out of whatever resource is being asked for
    SDIO_API_STATUS_PENDING,    //!< Unable to complete immediately
    SDIO_API_STATUS_FAILURE,    //!< Function call unsuccessful
    SDIO_API_STATUS_INVALID_ARG,        //!< Invalid argument passed by the caller
    SDIO_API_STATUS_ERROR_UNKNOWN,      //!< Unknown system error occurred
    SDIO_API_STATUS_HWINIT_FAIL,        //!< Hardware failed to initialize
    SDIO_API_STATUS_BOOT_READY, //!< boot loader ready
    SDIO_API_STATUS_HOST_BOOT_READY,

    /* Add new entries above this */
    SDIO_API_STATUS_INVALID_STATUS_VALUE        // Anything equal or above this is invalid status.
} SDIO_API_STATUS;

#define	WLAN_UPLD_SIZE		2312

/** \def DEFINE_EXTERN_GUID(GUID_MRVL_SDIOBUS_INTERFACE_STANDARD)
* Define an Interface Guid to access the proprietary Marvell SDIO bus driver interface.
* This guid is used to identify a specific interface in IRP_MN_QUERY_INTERFACE
* handler.
*/

DEFINE_EXTERN_GUID(GUID_MRVL_SDIOBUS_INTERFACE_STANDARD);
// {B92D63EC-F143-4711-9853-DE24D1B85843}

/** \def DEFINE_EXTERN_GUID(GUID_MRVL_SDIOBUS_BUS_TYPE)
* Define a Guid for SDIO bus type. This is returned in response to
* IRP_MN_QUERY_BUS_INTERFACE on PDO.
*/
DEFINE_EXTERN_GUID(GUID_MRVL_SDIOBUS_BUS_TYPE);
// {43159826-A758-4aa2-9F98-AB229C031ED1}

/** \def DEFINE_EXTERN_GUID(GUID_MRVL_SDIOBUS_INTERFACE_2_STANDARD)
* Define an Interface Guid to access the proprietary Marvell SDIO bus driver interface.
* This guid is used to identify a specific interface in IRP_MN_QUERY_INTERFACE
* handler.
*/
// {1D5837EE-F656-46b4-A8DA-5484A25AB13A}
DEFINE_EXTERN_GUID(GUID_MRVL_SDIOBUS_INTERFACE_2_STANDARD);

/** \def DEFINE_EXTERN_GUID(GUID_MRVL_SDIOBUS_BUS_2_TYPE)
* Define a Guid for SDIO bus type. This is returned in response to
* IRP_MN_QUERY_BUS_INTERFACE on PDO.
*/
// {8C3E2C85-D056-4689-B678-83A2E6FADCF9}
DEFINE_EXTERN_GUID(GUID_MRVL_SDIOBUS_BUS_2_TYPE);

// Definitions for IOCTL handling

#ifdef WINNT

#define FILE_DEVICE_SDIOBUSENUM         FILE_DEVICE_BUS_EXTENDER
#define SDIOBUSENUM_IOCTL(_index_) \
    CTL_CODE (FILE_DEVICE_SDIOBUSENUM, _index_, METHOD_BUFFERED, FILE_ANY_ACCESS)

#else // #ifdef WINNT

#define FILE_DEVICE_SDIOBUSENUM 0x2A    // same as for Windows
#define SDIOBUSENUM_IOCTL(_index_) \
		_IOWR(FILE_DEVICE_SDIOBUSENUM, _index_, unsigned int )

#endif

#define IOCTL_SDIOBUSENUM_PLUGIN_HARDWARE               SDIOBUSENUM_IOCTL (0x0)
#define IOCTL_SDIOBUSENUM_UNPLUG_HARDWARE               SDIOBUSENUM_IOCTL (0x1)
#define IOCTL_SDIOBUSENUM_EJECT_HARDWARE                SDIOBUSENUM_IOCTL (0x2)
#define IOCTL_SDIOBUS_DONT_DISPLAY_IN_UI_DEVICE         SDIOBUSENUM_IOCTL (0x3)
// IOCTL for FPGA Configuration GUI
#define IOCTL_SDIOBUSENUM_SET_VOLTAGE                   SDIOBUSENUM_IOCTL (0x4)
#define IOCTL_SDIOBUSENUM_SET_BUSTYPE                   SDIOBUSENUM_IOCTL (0x5)
#define IOCTL_SDIOBUSENUM_SET_BUSWIDTH                  SDIOBUSENUM_IOCTL (0x6)
#define IOCTL_SDIOBUSENUM_SET_CLKSPEED                  SDIOBUSENUM_IOCTL (0x7)
#define IOCTL_SDIOBUSENUM_GET_BOARDREV                  SDIOBUSENUM_IOCTL (0x8)
#define IOCTL_SDIOBUSENUM_GET_JUMPER                    SDIOBUSENUM_IOCTL (0x9)

//
// Define Interface reference/dereference routines for
//  Interfaces exported by IRP_MN_QUERY_INTERFACE
//

typedef VOID(*PINTERFACE_REFERENCE) (PVOID Context);
typedef VOID(*PINTERFACE_DEREFERENCE) (PVOID Context);

typedef
    BOOLEAN
    (*PSDIO_CMD52_READ) (IN PVOID Context,
                         IN ULONG Offset,
                         IN UCHAR function_number, OUT UCHAR * pReturnData);
typedef
    BOOLEAN
    (*PSDIO_CMD52_WRITE) (IN PVOID Context,
                          IN ULONG Offset,
                          IN UCHAR function_number, IN UCHAR Data);

typedef
    BOOLEAN
    (*PSDIO_CMD53_READ) (IN PVOID Context,
                         IN ULONG Offset,
                         IN UCHAR function_number,
                         IN UCHAR mode,
                         IN UCHAR opcode, OUT UCHAR * pData, IN ULONG Count);
typedef
    BOOLEAN
    (*PSDIO_CMD53_WRITE) (IN PVOID Context,
                          IN ULONG Offset,
                          IN UCHAR function_number,
                          IN UCHAR mode,
                          IN UCHAR opcode, OUT UCHAR * pData, IN ULONG Count);

typedef
    BOOLEAN
    (*PSDIO_CMD53_READ_EX) (IN PVOID Context,
                            IN ULONG Offset,
                            IN UCHAR function_number,
                            IN UCHAR mode,
                            IN UCHAR opcode,
                            OUT UCHAR * pData,
                            IN ULONG Count, IN ULONG blksz);
typedef
    BOOLEAN
    (*PSDIO_CMD53_WRITE_EX) (IN PVOID Context,
                             IN ULONG Offset,
                             IN UCHAR function_number,
                             IN UCHAR mode,
                             IN UCHAR opcode,
                             OUT UCHAR * pData,
                             IN ULONG Count, IN ULONG blksz);

                                                                                                                                                                        /*--- mmoser 3/20/2006 ---*/
/*
 * reverse args order for compatibility of Marvell Linux drive
 */

typedef
    VOID
    (*PSDIO_ISR_HANDLER) (IN ULONG irq_status,
                          IN PVOID Context, IN PVOID regs);

typedef
    VOID
    (*PSDIO_REGISTER_ISR_HANDLER) (IN PVOID Context,
                                   IN PVOID isrContext,
                                   IN PSDIO_ISR_HANDLER pIsrHandler);

typedef VOID(*PSDIO_ENABLE_IRQ) (IN PVOID Context);
typedef VOID(*PSDIO_DISABLE_IRQ) (IN PVOID Context);

typedef BOOLEAN(*PSDIO_IS_CARD_PRESENT) (IN PVOID Context);

typedef BOOLEAN(*PSDIO_CREATE_CARD_THREAD) (IN PVOID Context);
typedef BOOLEAN(*PSDIO_DESTROY_CARD_THREAD) (IN PVOID Context);

typedef BOOLEAN(*PSDIO_SDIO_LOCKINTERFACE) (IN PVOID Context);
typedef VOID(*PSDIO_SDIO_UNLOCKINTERFACE) (IN PVOID Context);

typedef VOID(*PSDIO_SET_LED) (IN PVOID Context, IN UCHAR val, IN UCHAR on);
typedef VOID(*PSDIO_PS_STATE) (IN PVOID Context, IN UCHAR State);

typedef VOID(*PSDIO_SET_CLOCK_SPEED) (IN PVOID Context, IN ULONG bus_freq);

typedef VOID(*PSDIO_SET_CLOCK) (IN PVOID Context, IN UCHAR on);

typedef VOID(*PSDIO_SET_GPO) (IN PVOID Context, IN UCHAR on);

typedef VOID(*PSDIO_SET_BUSWIDTH) (IN PVOID Context, IN UCHAR width);

/** \struct MRVL_SDIO_INTERFACE_STANDARD
* \brief The Marvell SDIO Standard Interface to the bus driver
*/
typedef struct _MRVL_SDIO_INTERFACE_STANDARD
{
    USHORT Size;                                                                                                        /**< Size of this interface structure                   */
    USHORT Version;                                         /**< Version of the interface                           */
    PINTERFACE_REFERENCE InterfaceReference;                /**< Increments the interface reference counter         */
    PINTERFACE_DEREFERENCE InterfaceDereference;            /**< Decrements the interface reference counter         */
    PVOID Context;                                          /**< Pointer to a caller supplied structure             */
    PSDIO_CMD52_READ SDIO_Cmd52_Read;                                                                           /**< Pointer to CMD 52 Read function                    */
    PSDIO_CMD52_WRITE SDIO_Cmd52_Write;                                                                 /**< Pointer to CMD 52 Write function                   */
    PSDIO_CMD53_READ SDIO_Cmd53_Read;                                                                           /**< Pointer to CMD 53 Read function                    */
    PSDIO_CMD53_WRITE SDIO_Cmd53_Write;                                                                 /**< Pointer to CMD 53 Write function                   */
    PSDIO_CMD53_READ_EX SDIO_Cmd53_Read_Ex;                                                             /**< Pointer to CMD 53 extended Read function           */
    PSDIO_CMD53_WRITE_EX SDIO_Cmd53_Write_Ex;                                                           /**< Pointer to CMD 53 extended Write function          */
    PSDIO_REGISTER_ISR_HANDLER SDIO_RegisterIsrHandler;                         /**< Pointer to register interrupt handler function     */
    PSDIO_ENABLE_IRQ SDIO_EnableInterrupt;                                                                      /**< Pointer to enable FPGA interrupt function          */
    PSDIO_DISABLE_IRQ SDIO_DisableInterrupt;                                                            /**< Pointer to disable FPGA interrupt function         */
    PSDIO_IS_CARD_PRESENT SDIO_isCardPresent;                                                   /**< Pointer to check card present function             */
    PSDIO_SDIO_LOCKINTERFACE SDIO_lockInterface;                                        /**< Pointer to lock SDIO physical interface function   */
    PSDIO_SDIO_UNLOCKINTERFACE SDIO_unlockInterface;                            /**< Pointer to unlock SDIO physical interface function */
    PSDIO_SET_LED SDIO_SetLed;                                                                                          /**< Pointer to turn FPGA LED on/off function           */
    //  pweber 17.08.2005
    PSDIO_PS_STATE SDIO_PsState;                                                                                        /**< Pointer to set the power save state function       */

                                                                                                                                                                                                                                                                                                                                                                                        /*--- mmoser 12/18/2006 ---*/
    PSDIO_SET_CLOCK SDIO_SetClock;                                                                                                              /**< Pointer to set clock on/off function								*/
    PSDIO_SET_BUSWIDTH SDIO_SetBusWidth;                                                                                /**< Pointer to set bus width function  								*/
                                                                                                                                                                                                                                                /*--- mmoser 1/2/2007 ---*/
    PSDIO_SET_CLOCK_SPEED SDIO_SetClockSpeed;                                                                   /**< Pointer to set clock speed							*/

                                                                                                                                                                                                                                                                                                                                                                                /*--- mmoser 2/20/2007 ---*/
    PSDIO_SET_GPO SDIO_SetGPO;                                                                                                                          /**< Pointer to set general purpose output function			*/
} MRVL_SDIO_INTERFACE_STANDARD, *PMRVL_SDIO_INTERFACE_STANDARD;

// SDIO Interface macros

#define SDIO_CMD52_READ(pSdio,offset,fn_nr, pBuf) \
		(*(pSdio)->sdio_if.SDIO_Cmd52_Read)((pSdio)->sdio_if.Context,(offset),(fn_nr), (pBuf))

#define SDIO_CMD52_WRITE(pSdio,offset,fn_nr, data) \
		(*(pSdio)->sdio_if.SDIO_Cmd52_Write)((pSdio)->sdio_if.Context,(offset),(fn_nr), (data))

#define SDIO_CMD53_READ(pSdio,offset,fn_nr,mode,opcode,pBuf,length) \
		(*(pSdio)->sdio_if.SDIO_Cmd53_Read)((pSdio)->sdio_if.Context,(offset),(fn_nr),(mode),(opcode),(pBuf),(length))

#define SDIO_CMD53_WRITE(pSdio,offset,fn_nr,mode,opcode,pBuf,length) \
		(*(pSdio)->sdio_if.SDIO_Cmd53_Write)((pSdio)->sdio_if.Context,(offset),(fn_nr),(mode),(opcode),(pBuf),(length))

#define SDIO_CMD53_READ_EX(pSdio,offset,fn_nr,mode,opcode,pBuf,length,blksz) \
		(*(pSdio)->sdio_if.SDIO_Cmd53_Read_Ex)((pSdio)->sdio_if.Context,(offset),(fn_nr),(mode),(opcode),(pBuf),(length),(blksz))

#define SDIO_CMD53_WRITE_EX(pSdio,offset,fn_nr,mode,opcode,pBuf,length,blksz) \
		(*(pSdio)->sdio_if.SDIO_Cmd53_Write_Ex)((pSdio)->sdio_if.Context,(offset),(fn_nr),(mode),(opcode),(pBuf),(length),(blksz))

#define SDIO_IS_CARD_PRESENT(pSdio) \
	(*(pSdio)->sdio_if.SDIO_isCardPresent)((pSdio)->sdio_if.Context)

#define SDIO_SDIO_LOCKINTERFACE(pSdio) \
	(*(pSdio)->sdio_if.SDIO_lockInterface)((pSdio)->sdio_if.Context)

#define SDIO_SDIO_UNLOCKINTERFACE(pSdio) \
	(*(pSdio)->sdio_if.SDIO_unlockInterface)((pSdio)->sdio_if.Context)

#define SDIO_SET_LED(pSdio, led, on ) \
	(*(pSdio)->sdio_if.SDIO_SetLed)((pSdio)->sdio_if.Context,(led),(on))

//  pweber 17.08.2005
#define SDIO_PS_STATE(pSdio, State ) \
	(*(pSdio)->sdio_if.SDIO_PsState)((pSdio)->sdio_if.Context,(State) )

#define SDIO_ENABLE_IRQ(pSdio ) \
	(*(pSdio)->sdio_if.SDIO_EnableInterrupt)((pSdio)->sdio_if.Context )

#define SDIO_DISABLE_IRQ(pSdio ) \
	(*(pSdio)->sdio_if.SDIO_DisableInterrupt)((pSdio)->sdio_if.Context )

#define SDIO_REGISTER_ISR_HANDLER(pSdio, isrContext, pIsrHandler ) \
	(*(pSdio)->sdio_if.SDIO_RegisterIsrHandler)((pSdio)->sdio_if.Context,(isrContext),(pIsrHandler))

#define SDIO_SET_CLOCK(pSdio, on ) \
	(*(pSdio)->sdio_if.SDIO_SetClock)((pSdio)->sdio_if.Context,(on))

#define SDIO_SET_CLOCK_SPEED(pSdio, bus_freq ) \
	(*(pSdio)->sdio_if.SDIO_SetClockSpeed)((pSdio)->sdio_if.Context,(bus_freq))

#define SDIO_SET_BUSWIDTH(pSdio, width ) \
	(*(pSdio)->sdio_if.SDIO_SetBusWidth)((pSdio)->sdio_if.Context,(width))

#define SDIO_SET_GPO(pSdio, on ) \
	(*(pSdio)->sdio_if.SDIO_SetGPO)((pSdio)->sdio_if.Context,(on))

#define MAX_WLAN_TX_SIZE  3000
#define MAX_WLAN_RX_SIZE  3000
#define MAX_WLAN_RSP_SIZE  256
#define MAX_FUNCTIONS			2

#ifndef WINNT
                                                                                                                                                                        /*--- mmoser 3/20/2006 ---*/

#define sdio_write_ioreg(card, fn, reg, val)  (sdio_Write_Byte(((mmc_card_t)(card))->ctrlr,(reg), (fn), (val)) ? 0 : -1 )
#define sdio_read_ioreg( card, fn, reg, val)  (sdio_Read_Byte(((mmc_card_t)(card))->ctrlr,(reg), (fn), (val)) ? 0 : -1 )
#define sdio_write_iomem(card,fn,reg,blockmode,opcode,cnt,blksz,dat) \
 (sdio_Write_BlockEx(((mmc_card_t)(card))->ctrlr,reg,fn,blockmode,opcode,dat,cnt,blksz)? 0 : -1)
#define sdio_read_iomem(card,fn,reg,blockmode,opcode,cnt,blksz,dat) \
 (sdio_Read_BlockEx(((mmc_card_t)(card))->ctrlr,reg,fn,blockmode,opcode,dat,cnt,blksz)? 0 : -1)

#define sdio_request_irq(card, handler, flags,devname,dev_id) sdio_Register_ISR_Handler(((mmc_card_t)(card))->ctrlr,dev_id,(PSDIO_ISR_HANDLER)handler)
// #define sdio_free_irq(card, dev_id)  sdio_Register_ISR_Handler(((mmc_card_t)(card))->ctrlr,dev_id,NULL)
#define sdio_enable_SDIO_INT() sdio_EnableInterrupt(((mmc_card_t)((wlan_private *)priv)->wlan_dev.card)->ctrlr)

void sdio_card_added(void *);
void sdio_card_removed(void);

typedef struct _card_capability
{
    u8 num_of_io_funcs;         /* Number of i/o functions */
    u8 memory_yes;              /* Memory present ? */
    u16 rca;                    /* Relative Card Address */
    u32 ocr;                    /* Operation Condition register */
    u16 fnblksz[8];
    u32 cisptr[8];
} card_capability;

typedef struct _dummy_tmpl
{
    int irq_line;
} dummy_tmpl;

typedef struct _mmc_card_rec
{
    u8 chiprev;
    u8 block_size_512;
    u8 async_int_mode;
    struct device *dev;
    card_capability info;
    struct _SDIO_OBJECT *ctrlr;
} mmc_card_rec;

typedef struct _mmc_card_rec *mmc_card_t;

typedef int (*mmc_notifier_fn_t) (mmc_card_t);

typedef struct _mmc_notifier_rec mmc_notifier_rec_t;
typedef struct _mmc_notifier_rec *mmc_notifier_t;

typedef struct _SDIO_OBJECT *mmc_controller_t;

struct _mmc_notifier_rec
{
    mmc_notifier_fn_t add;
    mmc_notifier_fn_t remove;
};
#endif // WINNT

/** \struct SDIO_OBJECT
* \brief Structure of the SDIO object
*/
typedef struct _SDIO_OBJECT
{
    MRVL_SDIO_INTERFACE_STANDARD sdio_if;       //!< Standard interface to bus driver

    ULONG ioport;               //!< Wireless SDIO port address
    ULONG cisptr[MAX_FUNCTIONS];        //!< Pointer to CIS
    UCHAR pUpl_Buf[WLAN_UPLD_SIZE];     //!< Buffer for data from card (upload)
    UCHAR pDnl_Buf[WLAN_UPLD_SIZE];     //!< Buffer for data to card (download)
#ifdef WINNT
    KTIMER SleepTimer;          //!< \if .WINDOWS. Timer used in the _SLEEP_MS() macro \endif
#endif
                                                                                                                                                                        /*--- mmoser 3/3/2006 ---*/
#ifndef WINNT
    spinlock_t IFSpinLock;      //!< \if .LINUX. Interface spin lock \endif
    spinlock_t fwSpinLock;      //!< \if .LINUX. Firmware spin lock \endif

                                                                                                                                                                        /*--- mmoser 3/20/2006 ---*/
    struct _dummy_tmpl *tmpl;
    mmc_card_t card;
#endif
} SDIO_OBJECT, *PSDIO_OBJECT;

#ifndef WINNT

// bhaake
// #define EXTERN_C

#define SDIO_API_IMPORT
#define PDEVICE_OBJECT	PVOID

#endif

EXTERN_C SDIO_API_IMPORT SDIO_API_STATUS sdio_CreateIface(PSDIO_OBJECT *
                                                          ppSdio_if,
                                                          PDEVICE_OBJECT
                                                          pSdioPDO,
                                                          PVOID irq_context,
                                                          PSDIO_ISR_HANDLER
                                                          irq_handler);
EXTERN_C SDIO_API_IMPORT SDIO_API_STATUS sdio_DestroyIface(PSDIO_OBJECT *
                                                           ppSdio_if);
EXTERN_C SDIO_API_IMPORT BOOLEAN sdio_Read_Byte(PSDIO_OBJECT pSdio,
                                                ULONG Offset,
                                                UCHAR function_number,
                                                UCHAR * pData);
EXTERN_C SDIO_API_IMPORT BOOLEAN sdio_Write_Byte(PSDIO_OBJECT pSdio,
                                                 ULONG Offset,
                                                 UCHAR function_number,
                                                 UCHAR Data);
EXTERN_C SDIO_API_IMPORT BOOLEAN sdio_Read_Block(PSDIO_OBJECT pSdio,
                                                 ULONG Offset,
                                                 UCHAR function_number,
                                                 UCHAR mode, UCHAR opcode,
                                                 UCHAR * pData, ULONG Count);
EXTERN_C SDIO_API_IMPORT BOOLEAN sdio_Write_Block(PSDIO_OBJECT pSdio,
                                                  ULONG Offset,
                                                  UCHAR function_number,
                                                  UCHAR mode, UCHAR opcode,
                                                  UCHAR * pData, ULONG Count);

                                                                                                                                                                        /*--- mmoser 3/20/2006 ---*/
EXTERN_C SDIO_API_IMPORT BOOLEAN sdio_Read_BlockEx(PSDIO_OBJECT pSdio,
                                                   ULONG Offset,
                                                   UCHAR function_number,
                                                   UCHAR mode, UCHAR opcode,
                                                   UCHAR * pData, ULONG Count,
                                                   ULONG blksz);
EXTERN_C SDIO_API_IMPORT BOOLEAN sdio_Write_BlockEx(PSDIO_OBJECT pSdio,
                                                    ULONG Offset,
                                                    UCHAR function_number,
                                                    UCHAR mode, UCHAR opcode,
                                                    UCHAR * pData,
                                                    ULONG Count, ULONG blksz);

EXTERN_C SDIO_API_IMPORT BOOLEAN sdio_isCardPresent(PSDIO_OBJECT pSdio_if);
EXTERN_C SDIO_API_IMPORT BOOLEAN sdio_lockInterface(PSDIO_OBJECT pSdio);
EXTERN_C SDIO_API_IMPORT VOID sdio_unlockInterface(PSDIO_OBJECT pSdio);
EXTERN_C SDIO_API_IMPORT VOID sdio_SetLed(PSDIO_OBJECT pSdio, UCHAR led,
                                          UCHAR on);
EXTERN_C SDIO_API_IMPORT VOID sdio_PsState(PSDIO_OBJECT pSdio, UCHAR State);    //  pweber 17.08.2005
EXTERN_C SDIO_API_IMPORT VOID sdio_EnableInterrupt(PSDIO_OBJECT pSdio);
EXTERN_C SDIO_API_IMPORT VOID sdio_DisableInterrupt(PSDIO_OBJECT pSdio);
EXTERN_C SDIO_API_IMPORT ULONG sdio_Register_ISR_Handler(PSDIO_OBJECT pSdio,
                                                         PVOID isrContext,
                                                         PSDIO_ISR_HANDLER
                                                         pIsrHandler);
EXTERN_C SDIO_API_IMPORT VOID sdio_SetClock(PSDIO_OBJECT pSdio, UCHAR on);
EXTERN_C SDIO_API_IMPORT VOID sdio_SetGPO(PSDIO_OBJECT pSdio, UCHAR on);
EXTERN_C SDIO_API_IMPORT VOID sdio_SetClockSpeed(PSDIO_OBJECT pSdio,
                                                 ULONG bus_freq);
EXTERN_C SDIO_API_IMPORT VOID sdio_SetBusWidth(PSDIO_OBJECT pSdio,
                                               UCHAR width);

EXTERN_C SDIO_API_IMPORT BOOLEAN card_Helper_Download(PSDIO_OBJECT pSdio,
                                                      UCHAR * helperImage,
                                                      ULONG helpersize);
EXTERN_C SDIO_API_IMPORT SDIO_API_STATUS card_Firmware_Download(PSDIO_OBJECT
                                                                pSdio,
                                                                UCHAR *
                                                                pFW_Image,
                                                                ULONG fmsize,
                                                                UCHAR *
                                                                pFW_helperImage,
                                                                ULONG
                                                                helperfmsize);
EXTERN_C SDIO_API_IMPORT BOOLEAN card_isCardReady(PSDIO_OBJECT pSdio);
EXTERN_C SDIO_API_IMPORT BOOLEAN card_disable_card_int(PSDIO_OBJECT pSdio,
                                                       UCHAR function_number,
                                                       UCHAR mask);
EXTERN_C SDIO_API_IMPORT BOOLEAN card_enable_card_int(PSDIO_OBJECT pSdio,
                                                      UCHAR function_number,
                                                      UCHAR mask);

EXTERN_C SDIO_API_IMPORT SDIO_API_STATUS fw_CardToHost(PSDIO_OBJECT pSdio,
                                                       ULONG * type,
                                                       ULONG * length,
                                                       UCHAR * payload,
                                                       ULONG npayload);
EXTERN_C SDIO_API_IMPORT SDIO_API_STATUS fw_HostToCard(PSDIO_OBJECT pSdio,
                                                       UCHAR type,
                                                       UCHAR * payload,
                                                       ULONG npayload);

                                                                                                                                                                                                                                                                        /*--- mmoser 3/20/2006 ---*/
#ifndef WINNT
EXTERN_C SDIO_API_IMPORT VOID *sdio_register(int type, void *ops, int memory);
EXTERN_C SDIO_API_IMPORT int sdio_unregister(int type, void *ops);
EXTERN_C SDIO_API_IMPORT int sdio_get_vendor_id(mmc_card_t card);
EXTERN_C SDIO_API_IMPORT int start_bus_clock(mmc_controller_t ctrlr);
EXTERN_C SDIO_API_IMPORT int sdio_free_irq(mmc_card_t card, void *dev_id);

EXTERN_C SDIO_API_IMPORT void sdio_clear_imask(mmc_controller_t);
EXTERN_C SDIO_API_IMPORT int sdio_check_idle_state(mmc_controller_t);
EXTERN_C SDIO_API_IMPORT void sdio_print_imask(mmc_controller_t);
EXTERN_C SDIO_API_IMPORT void sdio_clear_imask(mmc_controller_t);

EXTERN_C SDIO_API_IMPORT void sdio_unmask(void);
#endif

#endif
