/******************************************************************************
 *
 * Name:	sdio_if.c
 * Project:	Wireless LAN, Interface to SDIO bus driver
 * Version:	$Revision: 1.1 $
 * Date:	$Date: 2007/01/18 09:30:19 $
 * Purpose:	SDIO interface to bus driver definitions
 *
 ******************************************************************************/

/******************************************************************************
 *
 *	(C)Copyright 2004 Marvell
 *
 *	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SYSKONNECT
 *	The copyright notice above does not evidence any
 *	actual or intended publication of such source code.
 *
 *	This Module contains Proprietary Information of SysKonnect
 *	and should be treated as Confidential.
 *
 *	The information in this file is provided for the exclusive use of
 *	the licensees of SysKonnect.
 *	Such users have the right to use, modify, and incorporate this code
 *	into products for purposes authorized by the license agreement
 *	provided they include this notice and the associated copyright notice
 *	with any such product.
 *	The information in this file is provided "AS IS" without warranty.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * History:
 *
 *	$Log: sdio_if.c,v $
 *	Revision 1.1  2007/01/18 09:30:19  pweber
 *	Put under CVS control
 *	
 *
 ******************************************************************************/
#define SDIO_BUS_DRIVER

#include "h/sdioapi.h"

// #define DEBUG_VERBOSE

#ifdef DEBUG_VERBOSE
ULONG j, k;
UCHAR __buffer[1024];
#endif

// Initialize the interface GUIDS
//
// Define an Interface Guid to access the proprietary Marvell SDIO bus driver interface.
// This guid is used to identify a specific interface in IRP_MN_QUERY_INTERFACE
// handler.
//
DEFINE_INIT_GUID(GUID_MRVL_SDIOBUS_INTERFACE_STANDARD,
                 0xB92D63EC, 0xF143, 0x4711, 0x98, 0x53, 0xDE, 0x24, 0xD1,
                 0xB8, 0x58, 0x43);
// {B92D63EC-F143-4711-9853-DE24D1B85843}

//
// Define a Guid for SDIO bus type. This is returned in response to
// IRP_MN_QUERY_BUS_INTERFACE on PDO.
//
DEFINE_INIT_GUID(GUID_MRVL_SDIOBUS_BUS_TYPE,
                 0x43159826, 0xA758, 0x4aa2, 0x9F, 0x98, 0xAB, 0x22, 0x9C,
                 0x03, 0x1E, 0xD1);
// {43159826-A758-4aa2-9F98-AB229C031ED1}

//
// Define an Interface Guid to access the proprietary Marvell SDIO bus driver interface.
// This guid is used to identify a specific interface in IRP_MN_QUERY_INTERFACE
// handler.
//
// {1D5837EE-F656-46b4-A8DA-5484A25AB13A}
DEFINE_INIT_GUID(GUID_MRVL_SDIOBUS_INTERFACE_2_STANDARD,
                 0x1d5837ee, 0xf656, 0x46b4, 0xa8, 0xda, 0x54, 0x84, 0xa2,
                 0x5a, 0xb1, 0x3a);

//
// Define a Guid for SDIO bus type. This is returned in response to
// IRP_MN_QUERY_BUS_INTERFACE on PDO.
//
// {8C3E2C85-D056-4689-B678-83A2E6FADCF9}
DEFINE_INIT_GUID(GUID_MRVL_SDIOBUS_BUS_2_TYPE,
                 0x8c3e2c85, 0xd056, 0x4689, 0xb6, 0x78, 0x83, 0xa2, 0xe6,
                 0xfa, 0xdc, 0xf9);

#ifndef WINNT
MODULE_DESCRIPTION("SDIO Bus Interface Driver Module");
MODULE_AUTHOR("Michael Moser (mmoser@syskonnect.de)");
MODULE_LICENSE("Dual BSD/GPL");
#endif

#define FWDELAY 2000            // cycles should give us about 10 seconds
#define FWBLOCK_DELAY  5000     // cycles for fw block delay.

// #define DEBUG_VERBOSE        1

#ifndef WINNT
int ps_sleep_confirmed = 0;

mmc_notifier_t pNotifier;

PSDIO_OBJECT global_pSdio = NULL;
#endif

                                                                                                                                                                        /*--- mmoser 3/31/2006 ---*/
static int card_is_removed = 1;
static int card_is_inserted = 0;

// private functions
ULONG card_GetCIS_Pointer(PSDIO_OBJECT pSdio, UCHAR function);
BOOLEAN card_ReadIO_Port(PSDIO_OBJECT pSdio);
UCHAR card_poll_card_status(PSDIO_OBJECT pSdio, UCHAR bits);
BOOLEAN card_read_scratch(PSDIO_OBJECT pSdio, ULONG * dat);
SDIO_API_STATUS sdio_GetStandardInterface(PVOID DevObject,
                                          PMRVL_SDIO_INTERFACE_STANDARD
                                          BusInterface);

/** 
 *  @brief Initialize the SDIO interface.
 *  				
 *  @param ppSdio_if  	Address of a pointer which returns the SDIO interface
 *  @param pSdioPDO			Pointer to the physical device object (only for Windows)
 *	@param irq_context	Pointer to a data structure which is passed to the interrupt handler
 *	@param irq_handler 	Pointer to the interrupt handler function
 * 
 *  @return 						SDIO_API_STATUS_FAILURE or SDIO_API_STATUS_SUCCESS
 */
SDIO_API_STATUS
sdio_CreateIface(PSDIO_OBJECT * ppSdio_if,
                 PDEVICE_OBJECT pSdioPDO,
                 PVOID irq_context, PSDIO_ISR_HANDLER irq_handler)
{
    PSDIO_OBJECT pSdio;
    SDIO_API_STATUS Status;

    DBGPRINT(DBG_HWIF, (KERN_DEBUG "sdio_CreateIface\n"));

#ifdef WINNT
    pSdio =
        (PSDIO_OBJECT) ExAllocatePoolWithTag(NonPagedPool,
                                             sizeof(SDIO_OBJECT), 'DSRM');
#else

    pSdio = (PSDIO_OBJECT) kmalloc(sizeof(SDIO_OBJECT), GFP_KERNEL);
#endif

    if (pSdio == NULL) {
        return SDIO_API_STATUS_FAILURE;
    }

#ifdef WINNT
    RtlZeroMemory(pSdio, sizeof(SDIO_OBJECT));
#else
    memset(pSdio, '\0', sizeof(SDIO_OBJECT));

                                                                                                                                                                                                                        /*--- mmoser 3/20/2006 ---*/
    pSdio->card = (mmc_card_t) kmalloc(sizeof(mmc_card_rec), GFP_KERNEL);
    if (pSdio->card == NULL) {
        kfree(pSdio);
        return SDIO_API_STATUS_FAILURE;
    }
    memset(pSdio->card, '\0', sizeof(mmc_card_t));
    pSdio->card->ctrlr = pSdio;

    if (!(pSdio->tmpl = kmalloc(sizeof(dummy_tmpl), GFP_ATOMIC))) {
        kfree(pSdio->card);
        kfree(pSdio);
        return SDIO_API_STATUS_FAILURE;
    }
#endif

    // Query the standard interface
    Status =
        sdio_GetStandardInterface(pSdioPDO,
                                  (PMRVL_SDIO_INTERFACE_STANDARD) pSdio);
    if (Status != SDIO_API_STATUS_SUCCESS) {
        *ppSdio_if = NULL;
#ifdef WINNT
        ExFreePool(pSdio);
#else

                                                                                                                                                                        /*--- mmoser 3/20/2006 ---*/
        kfree(pSdio->tmpl);
        kfree(pSdio->card);
        kfree(pSdio);
#endif
        return SDIO_API_STATUS_FAILURE;
    }

#ifdef WINNT
    KeInitializeTimer(&pSdio->SleepTimer);
#endif

                                                                                                                                                                                /*--- mmoser 3/3/2006 ---*/
#ifndef WINNT
    spin_lock_init(&pSdio->IFSpinLock);
    spin_lock_init(&pSdio->fwSpinLock);
#endif

    *ppSdio_if = pSdio;

                                                                                                                                                                        /*--- mmoser 4/4/2006 ---*/
#ifdef WINNT
    card_ReadIO_Port(pSdio);

    pSdio->cisptr[0] = card_GetCIS_Pointer(pSdio, 0);
    pSdio->cisptr[1] = card_GetCIS_Pointer(pSdio, 1);
#endif

    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG
              "sdio_CreateIface: ioport=0x%8.08X cisptr[0]=0x%8.08X cisptr[1]=0x%8.08X\n",
              pSdio->ioport, pSdio->cisptr[0], pSdio->cisptr[1]));

    (*pSdio->sdio_if.SDIO_RegisterIsrHandler) (pSdio->sdio_if.Context,
                                               irq_context, irq_handler);
    return SDIO_API_STATUS_SUCCESS;
}

/** 
 *  @brief Deferences the SDIO interface.
 *  				
 *  @param ppSdio_if  	Address of a pointer which returns the SDIO interface
 *
 *  @return 						SDIO_API_STATUS_FAILURE or SDIO_API_STATUS_SUCCESS
 */
SDIO_API_STATUS
sdio_DestroyIface(PSDIO_OBJECT * ppSdio_if)
{
    PSDIO_OBJECT pSdio;

    DBGPRINT(DBG_HWIF, (KERN_DEBUG "sdio_DestroyIface\n"));

    if (!*ppSdio_if) {
        DBGPRINT(DBG_ERROR,
                 (KERN_DEBUG
                  "Attempt to free interface with NULL POINTER\n"));
        return SDIO_API_STATUS_FAILURE;
    }
    pSdio = *ppSdio_if;

    pSdio->sdio_if.InterfaceDereference((PVOID) pSdio->sdio_if.Context);

#ifdef WINNT
    KeCancelTimer(&pSdio->SleepTimer);
#endif

#ifdef WINNT
    ExFreePool(*ppSdio_if);
#else
    kfree(*ppSdio_if);
#endif

    *ppSdio_if = NULL;
    return SDIO_API_STATUS_SUCCESS;
}

/** 
 *  @brief Read one byte from the SDIO device (CMD52)
 *  				
 *  @param pSdio						pointer to our interface object.
 *	@param Offset						Address within the SDIO card
 *	@param function_number	function number
 *	@param pReturnData      return data byte
 *
 *  @return 								TRUE or FALSE
 */
BOOLEAN
sdio_Read_Byte(PSDIO_OBJECT pSdio,      // pointer to our interface object.
               ULONG Offset,    // Address within the SDIO card
               UCHAR function_number,   // function number
               UCHAR * pReturnData      // return data byte
    )
{
    BOOLEAN rc;
    LOCK_IF();
    rc = SDIO_CMD52_READ(pSdio, Offset, function_number, pReturnData);
    UNLOCK_IF();
    return rc;
}

/** 
 *  @brief Write one byte to the SDIO device (CMD52)
 *  				
 *  @param pSdio						pointer to our interface object.
 *	@param Offset						Address within the SDIO card
 *	@param function_number	function number
 *	@param Data     	 			data byte
 *
 *  @return 								TRUE or FALSE
 */
BOOLEAN
sdio_Write_Byte(PSDIO_OBJECT pSdio,     // pointer to our interface object.
                ULONG Offset,   // Address within the SDIO card
                UCHAR function_number,  // function number
                UCHAR Data      // data byte
    )
{
    BOOLEAN rc;
    LOCK_IF();
    rc = SDIO_CMD52_WRITE(pSdio, Offset, function_number, Data);
    UNLOCK_IF();
    return rc;
}

/** 
 *  @brief Read 'Count' bytes from the SDIO device (CMD53)
 *  				
 *  @param pSdio						pointer to our interface object.
 *	@param Offset						Address within the SDIO card
 *	@param function_number	function number
 *	@param mode 						byte/block mode
 *	@param opcode 					fixed/incrementing address
 *	@param pData 			      return data buffer
 *	@param Count						size of data buffer
 *
 *  @return 								TRUE or FALSE
 */
BOOLEAN
sdio_Read_Block(PSDIO_OBJECT pSdio,     // pointer to our interface object.
                ULONG Offset,   // Address within the SDIO card
                UCHAR function_number,  // function number
                UCHAR mode,     // byte/block mode
                UCHAR opcode,   // fixed/incrementing address
                UCHAR * pData,  // return data buffer
                ULONG Count     // size of data buffer
    )
{
    BOOLEAN rc;
    LOCK_IF();
    rc = SDIO_CMD53_READ(pSdio, Offset, function_number, mode, opcode, pData,
                         Count);
    UNLOCK_IF();
    return rc;
}

/** 
 *  @brief Write 'Count' bytes to the SDIO device (CMD53)
 *  				
 *  @param pSdio						pointer to our interface object.
 *	@param Offset						Address within the SDIO card
 *	@param function_number	function number
 *	@param mode 						byte/block mode
 *	@param opcode 					fixed/incrementing address
 *	@param pData 			      return data buffer
 *	@param Count						size of data buffer
 *
 *  @return 								TRUE or FALSE
 */
BOOLEAN
sdio_Write_Block(PSDIO_OBJECT pSdio,    // pointer to our interface object.
                 ULONG Offset,  // Address within the SDIO card
                 UCHAR function_number, // function number
                 UCHAR mode,    // byte/block mode
                 UCHAR opcode,  // fixed/incrementing address
                 UCHAR * pData, // data buffer
                 ULONG Count    // size of data buffer
    )
{
    BOOLEAN rc;
    LOCK_IF();
    rc = SDIO_CMD53_WRITE(pSdio, Offset, function_number, mode, opcode, pData,
                          Count);
    UNLOCK_IF();
    return rc;
}

/** 
 *  @brief Read 'Count' bytes with block size 'blksz' from the SDIO device (CMD53)
 *  				
 *  @param pSdio						pointer to our interface object.
 *	@param Offset						Address within the SDIO card
 *	@param function_number	function number
 *	@param mode 						byte/block mode
 *	@param opcode 					fixed/incrementing address
 *	@param pData 			      return data buffer
 *	@param Count						size of data buffer
 *	@param blksz						block size
 *
 *  @return 								TRUE or FALSE
 */
BOOLEAN
sdio_Read_BlockEx(PSDIO_OBJECT pSdio,   // pointer to our interface object.
                  ULONG Offset, // Address within the SDIO card
                  UCHAR function_number,        // function number
                  UCHAR mode,   // byte/block mode
                  UCHAR opcode, // fixed/incrementing address
                  UCHAR * pData,        // return data buffer
                  ULONG Count,  // size of data buffer
                  ULONG blksz   // block size
    )
{
    BOOLEAN rc;
    LOCK_IF();
    rc = SDIO_CMD53_READ_EX(pSdio, Offset, function_number, mode, opcode,
                            pData, Count, blksz);
    UNLOCK_IF();
    return rc;
}

/** 
 *  @brief Write 'Count' bytes with block size 'blksz' to the SDIO device (CMD53)
 *  				
 *  @param pSdio						pointer to our interface object.
 *	@param Offset						Address within the SDIO card
 *	@param function_number	function number
 *	@param mode 						byte/block mode
 *	@param opcode 					fixed/incrementing address
 *	@param pData 			      return data buffer
 *	@param Count						size of data buffer
 *	@param blksz						block size
 *
 *  @return 								TRUE or FALSE
 */
BOOLEAN
sdio_Write_BlockEx(PSDIO_OBJECT pSdio,  // pointer to our interface object.
                   ULONG Offset,        // Address within the SDIO card
                   UCHAR function_number,       // function number
                   UCHAR mode,  // byte/block mode
                   UCHAR opcode,        // fixed/incrementing address
                   UCHAR * pData,       // data buffer
                   ULONG Count, // size of data buffer
                   ULONG blksz  // block size
    )
{
    BOOLEAN rc;
    LOCK_IF();
    rc = SDIO_CMD53_WRITE_EX(pSdio, Offset, function_number, mode, opcode,
                             pData, Count, blksz);
    UNLOCK_IF();
    return rc;
}

/** 
 *  @brief Fetch one firmware message from the SDIO device
 *  				
 *  @param pSdio						pointer to our interface object.
 * 	@param type							type of firmware message
 * 	@param nb								received message size
 * 	@param payload					returns the message to the caller
 * 	@param npayload					max. allowed payload size
 *
 *  @return 						SDIO_API_STATUS_FAILURE or SDIO_API_STATUS_SUCCESS
 */
SDIO_API_STATUS
fw_CardToHost(PSDIO_OBJECT pSdio, ULONG * type, ULONG * nb, UCHAR * payload,
              ULONG npayload)
{
    BOOLEAN status = FALSE;
    ULONG buf_len = 0;
    ULONG blks;
    LOCK_FW();

    if (!sdio_isCardPresent(pSdio)) {
        DBGPRINT(DBG_HWIF, (KERN_DEBUG "fw_CardToHost: No Card Present!\n"));
        UNLOCK_FW();
        return SDIO_API_STATUS_FAILURE;
    }

    if (payload == NULL || nb == NULL || type == NULL) {
        DBGPRINT(DBG_HWIF,
                 (KERN_DEBUG
                  "fw_CardToHost: Invalid parameters payload=0x%8.08X length=%d \n",
                  (ULONG) payload, *nb));
        UNLOCK_FW();
        return SDIO_API_STATUS_FAILURE;
    }

    /* Read the length of data to be transferred */
    status = card_read_scratch(pSdio, &buf_len);
    if (!status) {
        UNLOCK_FW();
        return SDIO_API_STATUS_FAILURE;
    }
#ifdef DEBUG_VERBOSE
    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG
              "fw_CardToHost: Receiving %d bytes from card at scratch reg value.\n",
              buf_len));
#endif
    if (buf_len > WLAN_UPLD_SIZE || buf_len == 0) {
        UNLOCK_FW();
        return SDIO_API_STATUS_FAILURE;
    }
#ifdef WINNT
    RtlZeroMemory(pSdio->pUpl_Buf, buf_len * sizeof(UCHAR));
#else
    memset(pSdio->pUpl_Buf, '\0', buf_len * sizeof(UCHAR));
#endif

    blks = (buf_len + SD_BLOCK_SIZE - 1) / SD_BLOCK_SIZE;
    status = sdio_Read_Block(pSdio, pSdio->ioport, 1,
                             BLOCK_MODE, FIXED_ADDRESS,
                             pSdio->pUpl_Buf, blks);

    if (status == TRUE) {
#ifdef DEBUG_VERBOSE
        DBGPRINT(DBG_HWIF, (KERN_DEBUG "CardToHost : len=%d\n", buf_len));
#ifdef WINNT
        PrintMacro("CardToHost : len=%d\n", buf_len);
        for (j = 0; j < buf_len && j < 64; j++) {
            DbgPrint("%2.02X ", pSdio->pUpl_Buf[j]);
            if ((j + 1) % SD_BLOCK_SIZE == 0) {
                DbgPrint("\n");
            }
        }
        DbgPrint("\n");
#else // #ifdef WINNT
        k = 0;
        for (j = 0; j < buf_len && j < 64; j++) {
            sprintf(&__buffer[k], "%2.02X ", pSdio->pUpl_Buf[j]);
            k += 3;
            if ((j + 1) % 32 == 0) {
                __buffer[k++] = '\n';
                __buffer[k++] = 0;
                DBGPRINT(DBG_HWIF, (KERN_DEBUG "%s\n", __buffer));
                k = 0;
            }
        }
        __buffer[k++] = '\n';
        __buffer[k++] = 0;
        DBGPRINT(DBG_HWIF, (KERN_DEBUG "%s\n", __buffer));
#endif
#endif

        *nb = *((PUCHAR) pSdio->pUpl_Buf + 0);
        *nb |= *((PUCHAR) pSdio->pUpl_Buf + 1) << 8;

        if (*nb < 4 || *nb > npayload) {
            DBGPRINT(DBG_HWIF,
                     (KERN_DEBUG "*nb=%d (%d) !!!!!!!\n", *nb, buf_len));

            UNLOCK_FW();
            return SDIO_API_STATUS_FAILURE;
        }
        *nb -= 4;
        *type = *((PUCHAR) pSdio->pUpl_Buf + 2);
        *type |= *((PUCHAR) pSdio->pUpl_Buf + 3) << 8;

        if (npayload >= buf_len) {
#ifdef WINNT
            RtlCopyMemory(payload, ((PUCHAR) pSdio->pUpl_Buf + 4), *nb);
#else
            memcpy(payload, ((PUCHAR) pSdio->pUpl_Buf + 4), *nb);
#endif
        } else {
            DBGPRINT(DBG_ERROR,
                     (KERN_DEBUG
                      "fw_CardToHost FAILURE : npayload < buf_len !!!!!!!\n"));
            UNLOCK_FW();
            return SDIO_API_STATUS_FAILURE;
        }

        UNLOCK_FW();
        return SDIO_API_STATUS_SUCCESS;
    }

    DBGPRINT(DBG_ERROR, (KERN_DEBUG "fw_CardToHost FAILURE !!!!!!!\n"));
    UNLOCK_FW();
    return SDIO_API_STATUS_FAILURE;
}

/** 
 *  @brief Send one firmware message to the SDIO device
 *  				
 *  @param pSdio						pointer to our interface object.
 * 	@param type							type of firmware message
 * 	@param payload					message 
 * 	@param nb								message size
 *
 *  @return 						SDIO_API_STATUS_FAILURE or SDIO_API_STATUS_SUCCESS
 */
SDIO_API_STATUS
fw_HostToCard(PSDIO_OBJECT pSdio, UCHAR type, UCHAR * payload, ULONG nb)
{
    ULONG blks;
    BOOLEAN status;
    LOCK_FW();

                                                                                                                                                                        /*--- mmoser 3/4/2006 ---*/
    sdio_DisableInterrupt(pSdio);
    if (!sdio_isCardPresent(pSdio)) {
        DBGPRINT(DBG_ERROR, (KERN_DEBUG "fw_HostToCard: No Card Present!\n"));
        UNLOCK_FW();
                                                                                                                                                                        /*--- mmoser 3/4/2006 ---*/
        sdio_EnableInterrupt(pSdio);
        return SDIO_API_STATUS_FAILURE;
    }

    if (payload == NULL || nb == 0 || nb + 4 > WLAN_UPLD_SIZE) {
        DBGPRINT(DBG_ERROR,
                 (KERN_DEBUG
                  "fw_HostToCard: payload (%8.08X) or length (%d) is invalid!\n",
                  (ULONG) payload, nb));
        UNLOCK_FW();
                                                                                                                                                                        /*--- mmoser 3/4/2006 ---*/
        sdio_EnableInterrupt(pSdio);
        return SDIO_API_STATUS_FAILURE;
    }
#ifdef DEBUG_VERBOSE
    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG "fw_HostToCard(0x%8.08X): length=%d\n",
              pSdio->ioport, nb));
#endif

    /* The host polls for the DN_LD_CARD_RDY and IO_READY bits */
// FW doesn't set the DN_LD_CARD_RDY correctly
// status =  card_poll_card_status(pSdio, IO_READY|DN_LD_CARD_RDY);

    status = card_poll_card_status(pSdio, IO_READY);

    if (status == 0) {
        DBGPRINT(DBG_ERROR, (KERN_DEBUG "fw_HostToCard failed\n"));
        UNLOCK_FW();
                                                                                                                                                                        /*--- mmoser 3/4/2006 ---*/
        sdio_EnableInterrupt(pSdio);
        return SDIO_API_STATUS_FAILURE;
    }

    pSdio->pDnl_Buf[0] = (UCHAR) ((nb + 4) & 0xff);
    pSdio->pDnl_Buf[1] = (UCHAR) (((nb + 4) >> 8) & 0xff);
    pSdio->pDnl_Buf[2] = (type == MVSD_DAT) ? 0 : 1;
    pSdio->pDnl_Buf[3] = 0;

    blks = ((nb + 4) + SD_BLOCK_SIZE - 1) / SD_BLOCK_SIZE;

    /* Transfer data to card */

#ifdef WINNT
    RtlCopyMemory(&pSdio->pDnl_Buf[4], payload, nb);
#else
    memcpy(&pSdio->pDnl_Buf[4], payload, nb);
#endif

#ifdef DEBUG_VERBOSE
#ifdef WINNT
    for (j = 0; j < nb + 4 && j < nb + 4 && j < 64; j++) {
        DbgPrint("%2.02X ", pSdio->pDnl_Buf[j]);

        if ((j + 1) % 32 == 0) {
            DbgPrint("\n");
        }
    }
    DbgPrint("\n");
#else // #ifdef WINNT
    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG
              "HostToCard : type=0x%2.02X%2.02X len=%d blocks=%d\n",
              pSdio->pDnl_Buf[3], pSdio->pDnl_Buf[2], nb, blks));
    k = 0;
    for (j = 0; j < nb + 4 && j < 64; j++) {
        sprintf(&__buffer[k], "%2.02X ", pSdio->pDnl_Buf[j]);
        k += 3;
        if ((j + 1) % 32 == 0) {
            __buffer[k++] = '\n';
            __buffer[k++] = 0;
            DBGPRINT(DBG_HWIF, (KERN_DEBUG "%s\n", __buffer));
            k = 0;
        }
    }
    __buffer[k++] = '\n';
    __buffer[k++] = 0;
    DBGPRINT(DBG_HWIF, (KERN_DEBUG "%s\n", __buffer));
#endif
#endif

    status =
        sdio_Write_Block(pSdio, pSdio->ioport, 1, BLOCK_MODE, FIXED_ADDRESS,
                         pSdio->pDnl_Buf, blks);

    if (!status) {
        DBGPRINT(DBG_ERROR,
                 (KERN_DEBUG
                  "fw_HostToCard(): IO error: transferring block\n"));
        UNLOCK_FW();
                                                                                                                                                                        /*--- mmoser 3/4/2006 ---*/
        sdio_EnableInterrupt(pSdio);
        return SDIO_API_STATUS_FAILURE;
    }
    UNLOCK_FW();
                                                                                                                                                                        /*--- mmoser 3/4/2006 ---*/
    sdio_EnableInterrupt(pSdio);
    return SDIO_API_STATUS_SUCCESS;
}

/** 
 *  @brief Check whether a SDIO device is present
 *  				
 *  @param pSdio		pointer to our interface object.
 *
 *  @return 				TRUE or FALSE
 */
BOOLEAN
sdio_isCardPresent(PSDIO_OBJECT pSdio)
{
    return SDIO_IS_CARD_PRESENT(pSdio);
}

/** 
 *  @brief Check whether the SDIO device is ready
 *  				
 *  @param pSdio		pointer to our interface object.
 *
 *  @return 				TRUE or FALSE
 */
BOOLEAN
card_isCardReady(PSDIO_OBJECT pSdio)
{
    UCHAR cs;

    if (!sdio_Read_Byte(pSdio, CARD_STATUS_REG, 1, &cs)) {
        return FALSE;
    }

    return (cs & DN_LD_CARD_RDY) && (cs & IO_READY);
}

/** 
 *  @brief Lock the SDIO interface
 *  				
 *  @param pSdio		pointer to our interface object.
 *
 *  @return 				TRUE or FALSE
 */
BOOLEAN
sdio_lockInterface(PSDIO_OBJECT pSdio)
{
    return SDIO_SDIO_LOCKINTERFACE(pSdio);
}

/** 
 *  @brief Unlock the SDIO interface
 *  				
 *  @param pSdio		pointer to our interface object.
 *
 */
VOID
sdio_unlockInterface(PSDIO_OBJECT pSdio)
{
    SDIO_SDIO_UNLOCKINTERFACE(pSdio);
}

/** 
 *  @brief Turn on/off the LED on the FPGA interface
 *  				
 *  @param pSdio		pointer to our interface object.
 *  @param led			LED number (0..4)
 *  @param on				on/off
 *
 */
VOID
sdio_SetLed(PSDIO_OBJECT pSdio, UCHAR led, UCHAR on)
{
    SDIO_SET_LED(pSdio, led, on);
}

/** 
 *  @brief Turn on/off the SDIO bus speed
 *  				
 *  @param pSdio		pointer to our interface object.
 *  @param bus_freq bus_freq
 *
 */
VOID
sdio_SetClockSpeed(PSDIO_OBJECT pSdio, ULONG bus_freq)
{
    SDIO_SET_CLOCK_SPEED(pSdio, bus_freq);
}

/** 
 *  @brief Turn on/off the SDIO bus clock
 *  				
 *  @param pSdio		pointer to our interface object.
 *  @param on				on/off
 *
 */
VOID
sdio_SetClock(PSDIO_OBJECT pSdio, UCHAR on)
{
    SDIO_SET_CLOCK(pSdio, on);
}

/** 
 *  @brief Turn on/off the general purpose output clock
 *  				
 *  @param pSdio		pointer to our interface object.
 *  @param on				on/off
 *
 */
VOID
sdio_SetGPO(PSDIO_OBJECT pSdio, UCHAR on)
{
    SDIO_SET_GPO(pSdio, on);
}

/** 
 *  @brief Set the SDIO bus width to 1 or 4 bit
 *  				
 *  @param pSdio		pointer to our interface object.
 *  @param on				on/off
 *
 */
VOID
sdio_SetBusWidth(PSDIO_OBJECT pSdio, UCHAR width)
{
    SDIO_SET_BUSWIDTH(pSdio, width);
}

//   pweber 17.08.2005
/** 
 *  @brief Set the power save state in the bus driver
 *  				
 *  @param pSdio		pointer to our interface object.
 *  @param State		power save state
 *
 */
VOID
sdio_PsState(PSDIO_OBJECT pSdio, UCHAR State)
{
    SDIO_PS_STATE(pSdio, State);
}

                                                                                                                                                                        /*--- mmoser 3/20/2006 ---*/
/** 
 *  @brief Register an interrupt call back in the bus driver
 *  				
 *  @param pSdio				pointer to our interface object.
 *	@param isrContext		Pointer to a data structure which is passed to the interrupt handler
 *	@param pIsrHandler 	Pointer to the interrupt handler function
 *
 */
ULONG
sdio_Register_ISR_Handler(PSDIO_OBJECT pSdio,
                          PVOID isrContext, PSDIO_ISR_HANDLER pIsrHandler)
{
    SDIO_REGISTER_ISR_HANDLER(pSdio, isrContext, pIsrHandler);
    return 0;
}

/** 
 *  @brief Enable the interrupt in the FPGA interface
 *  				
 *  @param pSdio		pointer to our interface object.
 *
 */
VOID
sdio_EnableInterrupt(PSDIO_OBJECT pSdio)
{
    SDIO_ENABLE_IRQ(pSdio);
}

/** 
 *  @brief Disable the interrupt in the FPGA interface
 *  				
 *  @param pSdio		pointer to our interface object.
 *
 */
VOID
sdio_DisableInterrupt(PSDIO_OBJECT pSdio)
{
    SDIO_DISABLE_IRQ(pSdio);
}

// private helper functions

/** 
 *  @brief Get the pointer to the CIS
 *  				
 *  @param pSdio		pointer to our interface object.
 * 	@param function	function number
 * 
 *	@return 				address of CIS
 *
 */
ULONG
card_GetCIS_Pointer(PSDIO_OBJECT pSdio, UCHAR function)
{
    UCHAR R;
    ULONG _cis = 0;

    if (!sdio_Read_Byte(pSdio, FN_CIS_POINTER_0_REG(function), 0, &R)) {
        return 0;
    }
    _cis = R;
    if (!sdio_Read_Byte(pSdio, FN_CIS_POINTER_1_REG(function), 0, &R)) {
        return 0;
    }
    _cis |= (R << 8);
    if (!sdio_Read_Byte(pSdio, FN_CIS_POINTER_2_REG(function), 0, &R)) {
        return 0;
    }
    _cis |= (R << 16);

    return _cis;
}

/** 
 *  @brief Get the IO port address for function 1
 *  				
 *  @param pSdio		pointer to our interface object.
 *
 *  @return 				TRUE or FALSE
 */
BOOLEAN
card_ReadIO_Port(PSDIO_OBJECT pSdio)
{
    UCHAR R;
    ULONG _port = 0;

    if (!sdio_Read_Byte(pSdio, IO_PORT_0_REG, 1, &R)) {
        return FALSE;
    }
    _port = R;
    if (!sdio_Read_Byte(pSdio, IO_PORT_1_REG, 1, &R)) {
        return FALSE;
    }
    _port |= (R << 8);
    if (!sdio_Read_Byte(pSdio, IO_PORT_2_REG, 1, &R)) {
        return FALSE;
    }
    _port |= (R << 16);

    pSdio->ioport = _port;

    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG "card_ReadIO_Port : IOPort=%8.08X\n", _port));

    return TRUE;
}

/** 
 *  @brief Download the helper firmware image to the SDIO device
 *  				
 *  @param pSdio				pointer to our interface object.
 *  @param helperImage	pointer to the helper firmware image
 * 	@param helpersize		size of the helper firmware image
 *
 *  @return 				TRUE or FALSE
 */
BOOLEAN
card_Helper_Download(PSDIO_OBJECT pSdio,
                     UCHAR * helperImage, ULONG helpersize)
{
    ULONG txed_len, tries, txlen;
    BOOLEAN status = FALSE;
    UCHAR fwbuf[FIRMWARE_TRANSFER_NBLOCK * SD_BLOCK_SIZE];      /*FIRMWARE_DOWNLOAD_BLOCK_SIZE */

    DBGPRINT(DBG_HWIF, (KERN_DEBUG "card_Helper_Download() \n"));

    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG "Downloading helper image (%d bytes)\n",
              helpersize));

    // Perform firmware data transfer
    txlen = (FIRMWARE_TRANSFER_NBLOCK * SD_BLOCK_SIZE) - 4;
    for (txed_len = 0; txed_len < helpersize; txed_len += txlen) {
        // The host polls for the DN_LD_CARD_RDY and IO_READY bits /
        status = card_poll_card_status(pSdio, IO_READY | DN_LD_CARD_RDY);

        if ((status & (IO_READY | DN_LD_CARD_RDY)) !=
            (IO_READY | DN_LD_CARD_RDY)) {
            DBGPRINT(DBG_HWIF,
                     (KERN_DEBUG "Helper image download died @ %d \r\n",
                      txed_len));
            return FALSE;
        }

        /* Set blocksize to transfer - checking for last block */
        if (helpersize - txed_len < txlen) {
            txlen = helpersize - txed_len;
        }

        fwbuf[0] = (UCHAR) ((txlen & 0x000000ff) >> 0); /* Little-endian */
        fwbuf[1] = (UCHAR) ((txlen & 0x0000ff00) >> 8);
        fwbuf[2] = (UCHAR) ((txlen & 0x00ff0000) >> 16);
        fwbuf[3] = (UCHAR) ((txlen & 0xff000000) >> 24);

        /* Copy payload to buffer */
#ifdef WINNT
        RtlCopyMemory(&fwbuf[4], &helperImage[txed_len], txlen);
#else
        memcpy(&fwbuf[4], &helperImage[txed_len], txlen);
#endif
        /* send data */
        status = sdio_Write_Block(pSdio, pSdio->ioport, 1,
                                  BLOCK_MODE,
                                  FIXED_ADDRESS,
                                  fwbuf, FIRMWARE_TRANSFER_NBLOCK);
        if (!status) {
            DBGPRINT(DBG_HWIF,
                     (KERN_DEBUG "IO error: transferring block @ %d\n",
                      txed_len));
            return FALSE;
        }
    }
    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG "\nHelper image download done (%d/%d bytes)\n",
              txed_len, helpersize));

    // The host polls for the DN_LD_CARD_RDY and IO_READY bits /
    status = card_poll_card_status(pSdio, IO_READY | DN_LD_CARD_RDY);

    //Write last EOF data
    DBGPRINT(DBG_HWIF, (KERN_DEBUG "Transferring EOF block\n"));

#ifdef WINNT
    RtlZeroMemory(fwbuf, sizeof(fwbuf));
#else
    memset(fwbuf, '\0', sizeof(fwbuf));
#endif

    status =
        sdio_Write_Block(pSdio, pSdio->ioport, 1, BLOCK_MODE, FIXED_ADDRESS,
                         fwbuf, FIRMWARE_TRANSFER_NBLOCK);

    if (!status) {
        DBGPRINT(DBG_HWIF,
                 (KERN_DEBUG "IO error in writing EOF firmware block\n"));
        return FALSE;
    }
    // Wait for firmware initialization event
    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG "Wait for firmware initialization event\n"));
    for (tries = 0; tries < MAX_FIRMWARE_POLL_TRIES; tries++) {
        _SLEEP_MS(pSdio, 10);
    }
    return TRUE;
}

/** 
 *  @brief Download the helper firmware image and the firmware image to the SDIO device
 *  				
 *  @param pSdio						pointer to our interface object.
 *  @param pFW_Image				pointer to the firmware image
 * 	@param fmsize						size of the firmware image
 *  @param pFW_helperImage	pointer to the helper firmware image
 * 	@param helperfmsize			size of the helper firmware image
 *
 *  @return 				TRUE or FALSE
 */
SDIO_API_STATUS
card_Firmware_Download(PSDIO_OBJECT pSdio,
                       UCHAR * pFW_Image, ULONG fmsize,
                       UCHAR * pFW_helperImage, ULONG helperfmsize)
{
    ULONG loopCount;
    UCHAR returndata;
    SDIO_API_STATUS status = SDIO_API_STATUS_FAILURE;
    UCHAR base0;
    UCHAR base1;
    ULONG timeout = 5000;       // As per Hedley
    USHORT len;
    ULONG firmwarestat = 0;
    ULONG txed_len, tries, txlen;
    ULONG tx_len = 0;
    ULONG fwblknow = 0;
    UCHAR fwbuf[FIRMWARE_TRANSFER_NBLOCK * 32 * 8 + 8]; /*FIRMWARE_DOWNLOAD_BLOCK_SIZE */

    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG "card_Firmware_Download(size=%d bytes) \n", fmsize));

    status = card_read_scratch(pSdio, &firmwarestat);
    if (!status) {
        printk("Cannot get fw status!\n");
        return SDIO_API_STATUS_FAILURE;
    }

    if (firmwarestat == FIRMWARE_READY) {
        printk("FW already loaded!\n");
        return SDIO_API_STATUS_SUCCESS;
    }

    loopCount = 0xff;
    while (loopCount != 0) {
        // need to check if FW is ready for download

        returndata = 0;
        if (!sdio_Read_Byte(pSdio, 0x20, 1, &returndata)) {
            return SDIO_API_STATUS_FAILURE;
        }
        if (!(returndata & 0x8) || !(returndata & 0x1)) {
            // Not ready for packet download
            loopCount--;
            UDELAY(20);         // wait 20 us
        } else {
            status = SDIO_API_STATUS_SUCCESS;
            break;
        }
    }

    if (status != SDIO_API_STATUS_SUCCESS) {
        printk("Bootloader not ready for download!\n");
        return SDIO_API_STATUS_FAILURE;
    }
// printk("%s - %s:%d\n",__FUNCTION__,__FILE__,__LINE__);
    //disable int
    card_disable_card_int(pSdio, 1, 0xff);

    sdio_DisableInterrupt(pSdio);

                                                                                                                                                                        /*--- mmoser 11.10.2005 ---*/
                                                                                                                                                                        /*--- mmoser 10/26/2005 ---*/
//      card_ReadIO_Port(pSdio); --> do it in sdio_CreateIface()

    if (pFW_helperImage != NULL) {
//              printk("Download Helper Image.\n");
        // download helper image first
        if (!
            (status =
             card_Helper_Download(pSdio, pFW_helperImage, helperfmsize))) {
            DBGPRINT(DBG_HWIF,
                     (KERN_DEBUG "Helper image download FAILED! \n"));
            return SDIO_API_STATUS_FAILURE;
        }
    }

    /* Wait initially for the first non-zero value */
    do {

        if (!sdio_Read_Byte(pSdio, HOST_F1_RD_BASE_0, 1, &base0)) {
            DBGPRINT(DBG_HWIF,
                     (KERN_DEBUG "Failed to read FN1 blocksize 0!\n"));
            return SDIO_API_STATUS_FAILURE;
        }
        if (!sdio_Read_Byte(pSdio, HOST_F1_RD_BASE_1, 1, &base1)) {
            DBGPRINT(DBG_HWIF,
                     (KERN_DEBUG "Failed to read FN1 blocksize 1!\n"));
            return SDIO_API_STATUS_FAILURE;
        }
        len = (((USHORT) base1) << 8) | base0;
        UDELAY(1000);           // wait 1000 us
    } while (!len && --timeout);

    if (!timeout) {
        DBGPRINT(DBG_HWIF, (KERN_DEBUG "Failed to read FN1 blocksize!\n"));
        return SDIO_API_STATUS_FAILURE;
    }
    len &= ~BIT_0;

    // The host polls for the DN_LD_CARD_RDY and IO_READY bits
    status =
        card_poll_card_status(pSdio, IO_READY | DN_LD_CARD_RDY /*0x05 */ );
    if ((status & (IO_READY | DN_LD_CARD_RDY)) != (IO_READY | DN_LD_CARD_RDY)) {
        DBGPRINT(DBG_HWIF,
                 (KERN_DEBUG "DN_LD_CARD_RDY and IO_READY not ready!\n"));
        return SDIO_API_STATUS_FAILURE;
    }

    // Perform firmware data transfer
    txlen = (SD_BLOCK_SIZE) - 4;
    for (txed_len = 0; txed_len < fmsize; txed_len += txlen) {
        txlen = len;

        /* Set blocksize to transfer - checking for last block */
        if (fmsize - txed_len < txlen) {
            txlen = fmsize - txed_len;
        }
//                      DBGPRINT(DBG_HWIF,( KERN_DEBUG "FW: txed_len=%d, txlen = 0x%04X(%d)\n", txed_len,txlen,txlen));

        tx_len = (FIRMWARE_TRANSFER_NBLOCK * SD_BLOCK_SIZE * 8);
        for (fwblknow = 0; fwblknow < txlen; fwblknow += tx_len) {
            // The host polls for the DN_LD_CARD_RDY and IO_READY bits /
            status = card_poll_card_status(pSdio, IO_READY | DN_LD_CARD_RDY);

            if ((status & (IO_READY | DN_LD_CARD_RDY)) !=
                (IO_READY | DN_LD_CARD_RDY)) {
                DBGPRINT(DBG_HWIF,
                         (KERN_DEBUG "Firmware download died @ %d \r\n",
                          txed_len));
                return SDIO_API_STATUS_FAILURE;
            }
            if (txlen - fwblknow < tx_len) {
                tx_len = txlen - fwblknow;
            }

            /* Copy payload to buffer */
#ifdef WINNT
            RtlCopyMemory(&fwbuf[0], &pFW_Image[txed_len + fwblknow], tx_len);
#else
            memcpy(&fwbuf[0], &pFW_Image[txed_len + fwblknow], tx_len);
#endif
            /* send data */
            status =
                sdio_Write_Block(pSdio, pSdio->ioport, 1, BLOCK_MODE,
                                 FIXED_ADDRESS, fwbuf,
                                 FIRMWARE_TRANSFER_NBLOCK * 8);

            if (!status) {
                DBGPRINT(DBG_HWIF,
                         (KERN_DEBUG "IO error: transferring block @ %d\n",
                          txed_len + fwblknow));
                return SDIO_API_STATUS_FAILURE;
            }
        }
        do {
                                                                                                                                                                                                /*--- mmoser 3/10/2006 ---*/
            UDELAY(10);         // wait 10 us
            if (!sdio_Read_Byte(pSdio, HOST_F1_CARD_RDY, 1, &base0)) {
                DBGPRINT(DBG_HWIF,
                         (KERN_DEBUG "Failed to read FN1 card ready !\n"));
                return SDIO_API_STATUS_FAILURE;
            }
        } while (!(base0 & 0x08) || !(base0 & 0x01));

        if (!sdio_Read_Byte(pSdio, HOST_F1_RD_BASE_0, 1, &base0)) {
            DBGPRINT(DBG_HWIF,
                     (KERN_DEBUG "Failed to read FN1 blocksize 0!\n"));
            return SDIO_API_STATUS_FAILURE;
        }
        if (!sdio_Read_Byte(pSdio, HOST_F1_RD_BASE_1, 1, &base1)) {
            DBGPRINT(DBG_HWIF,
                     (KERN_DEBUG "Failed to read FN1 blocksize 1!\n"));
            return SDIO_API_STATUS_FAILURE;
        }
        len = (((USHORT) base1) << 8) | base0;

        if (!len) {
            if (txed_len == fmsize) {
                break;
            }
            len = 0x10;
        }
        if (len & BIT_0) {
            DBGPRINT(DBG_HWIF,
                     (KERN_DEBUG
                      "CRC32 Error indicated by the helper: len=0x%04X(%d)\n",
                      len, len));
            len &= ~BIT_0;
            /* Setting this to 0 to resend from same offset */
            txlen = 0;
        }
    }

    // Wait for firmware initialization event
    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG "Wait for firmware initialization event\n"));
    for (tries = 0; tries < 10 * MAX_FIRMWARE_POLL_TRIES; tries++) {
        status = card_read_scratch(pSdio, &firmwarestat);
        if (!status) {
            _SLEEP_MS(pSdio, 10);
            continue;
        }

        if (firmwarestat == FIRMWARE_READY) {
            status = SDIO_API_STATUS_SUCCESS;
            break;
        } else {
            _SLEEP_MS(pSdio, 10);
            status = SDIO_API_STATUS_FAILURE;
        }
    }

    if (status == SDIO_API_STATUS_FAILURE) {
        DBGPRINT(DBG_HWIF,
                 (KERN_DEBUG
                  "Timeout waiting for firmware to become active\n"));
    } else {
        DBGPRINT(DBG_HWIF, (KERN_DEBUG "Firmware Active OK\n"));
    }

    card_enable_card_int(pSdio, 1, UP_LD_HOST_INT_MASK | DN_LD_HOST_INT_MASK);

    sdio_EnableInterrupt(pSdio);
    return status;
}

/** 
 *  @brief  Read from the special scratch 'port'.
 *  				
 *  @param pSdio				pointer to our interface object.
 *  @param dat					returns scratch register value
 *
 *  @return 				TRUE or FALSE
 */
BOOLEAN
card_read_scratch(PSDIO_OBJECT pSdio, ULONG * dat)
{

    BOOLEAN ret = FALSE;
    UCHAR r0, r1;

    ret = sdio_Read_Byte(pSdio, CARD_OCR_0_REG, 1, &r0);

    if (!ret)
        return ret;

    ret = sdio_Read_Byte(pSdio, CARD_OCR_1_REG, 1, &r1);
    if (!ret)
        return ret;

    *dat = (((USHORT) r1) << 8) | r0;

    return TRUE;
}

/** 
 *  @brief  Poll the Card Status register until the bits specified in the argument are turned on
 *  				
 *  @param pSdio		pointer to our interface object.
 *  @param bits			card status bits to wait for	
 *
 *  @return 				TRUE or FALSE
 */
UCHAR
card_poll_card_status(PSDIO_OBJECT pSdio, UCHAR bits)
{
    int tries;
    UCHAR cs;
    BOOLEAN status = FALSE;
    BOOLEAN status1 = FALSE;

    for (tries = 0; tries < MAX_POLL_TRIES * 100; tries++) {
        status1 = sdio_Read_Byte(pSdio, CARD_STATUS_REG, 1, &cs);
        if (status1) {
            if ((cs & bits) == bits) {
                status = TRUE;
                break;
            }
        } else {
            DBGPRINT(DBG_HWIF,
                     (KERN_DEBUG
                      "card_poll_card_status() FAILED !!!!! ( --> sdio_Read_Byte() failed )\n"));
            return 0;
        }

        UDELAY(10);
    }

    if (!status) {
        DBGPRINT(DBG_HWIF,
                 (KERN_DEBUG
                  "card_poll_card_status() FAILED !!!!! (cs=0x%2.02X)\n",
                  cs));
    }
    return cs;
}

/** 
 *  @brief Disable the interrupt of the SDIO device
 *  				
 *  @param pSdio						pointer to our interface object.
 * 	@param function_number	function number
 *  @param mask							interrupt mask
 * 
 *	@return		TRUE or FALSE
 */
BOOLEAN
card_disable_card_int(PSDIO_OBJECT pSdio, UCHAR function_number, UCHAR mask)
{
    BOOLEAN ret;
    UCHAR host_int_mask;
    UCHAR ucValue;

    /* Read back the host_int_mask register */
    sdio_Read_Byte(pSdio, HOST_INT_MASK_REG, function_number, &host_int_mask);
    ucValue = 0;
    /* Update with the mask and write back to the register */
    host_int_mask &= ~mask;
    ret =
        sdio_Write_Byte(pSdio, HOST_INT_MASK_REG, function_number,
                        host_int_mask);
    sdio_Write_Byte(pSdio, HOST_INT_MASK_REG, 0, host_int_mask);
    sdio_Write_Byte(pSdio, HOST_INTERRUPT_MASK_REG, 0, 0);

    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG
              "card_disable_card_int() : card:0x%2.02X bus:0x%2.02X\n",
              host_int_mask, ucValue));

    return ret;
}

/** 
 *  @brief Enable the interrupt of the SDIO device
 *  				
 *  @param pSdio						pointer to our interface object.
 * 	@param function_number	function number
 *  @param mask							interrupt mask
 * 
 *	@return		TRUE or FALSE
 */
BOOLEAN
card_enable_card_int(PSDIO_OBJECT pSdio, UCHAR function_number, UCHAR mask)
{
    BOOLEAN status;
    UCHAR ucValue;

    /* Simply write the mask to the register */
//      mask = 0x03;
    mask = 0x0F;

                                                                                                                                                                        /*--- mmoser 11.10.2005 ---*/
    sdio_Write_Byte(pSdio, HOST_INT_RSR_REG, 1,
                    UP_LD_HOST_INT_MASK | DN_LD_HOST_INT_MASK);

    status = sdio_Write_Byte(pSdio, HOST_INT_MASK_REG, function_number, mask);
    sdio_Write_Byte(pSdio, HOST_INTERRUPT_MASK_REG, 0, 0x03);

    ucValue = 0;

    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG
              "card_enable_card_int() : card:0x%2.02X bus:0x%2.02X\n", mask,
              ucValue));
    return status;
}

/** 
 *  @brief Get the bus interface standard information from the bus driver.
 *  				
 *  @param DevObject				pointer to the device object.
 * 	@param BusInterface			pointer to the bus interface
 * 
 *  @return 						SDIO_API_STATUS_FAILURE or SDIO_API_STATUS_SUCCESS
 */
SDIO_API_STATUS
sdio_GetStandardInterface(IN PVOID DevObject,
                          OUT PMRVL_SDIO_INTERFACE_STANDARD BusInterface)
{
#ifdef WINNT
    KEVENT event;
    NTSTATUS status;
    PIRP irp;
    IO_STATUS_BLOCK ioStatusBlock;
    PIO_STACK_LOCATION irpStack;
    PDEVICE_OBJECT targetObject;

    DBGPRINT(DBG_LOAD, (KERN_DEBUG "Init: sdio_GetStandardInterface\n"));

    KeInitializeEvent(&event, SynchronizationEvent, FALSE);

    targetObject = IoGetAttachedDeviceReference((PDEVICE_OBJECT) DevObject);

    irp = IoBuildSynchronousFsdRequest(IRP_MJ_PNP,
                                       targetObject,
                                       NULL, 0, NULL, &event, &ioStatusBlock);
    if (irp == NULL) {
        status = STATUS_INSUFFICIENT_RESOURCES;
        ObDereferenceObject(targetObject);
        DBGPRINT(DBG_LOAD,
                 (KERN_DEBUG
                  "Init: sdio_GetStandardInterface failed : insufficient resources\n"));
        return SDIO_API_STATUS_FAILURE;
    }

    irpStack = IoGetNextIrpStackLocation(irp);
    irpStack->MinorFunction = IRP_MN_QUERY_INTERFACE;
    irpStack->Parameters.QueryInterface.InterfaceType =
        (LPGUID) & GUID_MRVL_SDIOBUS_INTERFACE_2_STANDARD;
    irpStack->Parameters.QueryInterface.Size =
        sizeof(MRVL_SDIO_INTERFACE_STANDARD);
    irpStack->Parameters.QueryInterface.Version = 1;
    irpStack->Parameters.QueryInterface.Interface = (PINTERFACE) BusInterface;
    irpStack->Parameters.QueryInterface.InterfaceSpecificData = NULL;

    //
    // Initialize the status to error in case the bus driver decides not to
    // set it correctly.
    //

    irp->IoStatus.Status = STATUS_NOT_SUPPORTED;

    status = IoCallDriver(targetObject, irp);

    if (status == STATUS_PENDING) {
        LARGE_INTEGER Timeout;

        Timeout.QuadPart = -((LONGLONG) MILLI_SECOND * (LONGLONG) (2000));

        if (STATUS_SUCCESS !=
            KeWaitForSingleObject(&event, Executive, KernelMode, FALSE,
                                  &Timeout)) {
            // Failed !!!
            DBGPRINT(DBG_LOAD,
                     (KERN_DEBUG
                      "Init: sdio_GetStandardInterface failed : no answer from bus driver (timeout)\n"));
            ObDereferenceObject(targetObject);
            return SDIO_API_STATUS_FAILURE;
        }
        status = ioStatusBlock.Status;
    }

    if (STATUS_SUCCESS != status) {
        DBGPRINT(DBG_ERROR,
                 (KERN_DEBUG
                  "Init: sdio_GetStandardInterface failed : no answer from bus driver (status=%8.08X)\n",
                  status));
        ObDereferenceObject(targetObject);
        return SDIO_API_STATUS_FAILURE;
    }
    //
    // Done with reference
    //
    ObDereferenceObject(targetObject);
#else // #ifdef WINNT
    DBGPRINT(DBG_LOAD, (KERN_DEBUG "Init: sdio_GetStandardInterface\n"));

    if (mrvlsdio_get_interface
        ((PGUID) & GUID_MRVL_SDIOBUS_INTERFACE_2_STANDARD,
         sizeof(MRVL_SDIO_INTERFACE_STANDARD), 1, (PVOID) BusInterface) < 0) {
        DBGPRINT(DBG_ERROR,
                 (KERN_DEBUG "Init: sdio_GetStandardInterface failed\n"));
        return SDIO_API_STATUS_FAILURE;
    }
#endif
    return SDIO_API_STATUS_SUCCESS;
}

#ifndef WINNT

void
sdio_card_added(void *dev)
{
                                                                                                                                                                                /*--- mmoser 4/4/2006 ---*/
    if (pNotifier == NULL) {
        printk("%s: pNotifier == NULL!!!\n", __FUNCTION__);
        return;
    }
    DBGPRINT(DBG_HWIF,
             (KERN_DEBUG "%s called :pNotifier->add=%p global_pSdio=%p\n",
              __FUNCTION__, pNotifier->add, global_pSdio));
                                                                                                                                                        /*--- mmoser 3/31/2006 ---*/
    card_is_removed = 0;
    card_is_inserted = 1;
    if (pNotifier->add != NULL) {
        if (global_pSdio == NULL) {
            printk("%s: global_pSdio == NULL!!!\n", __FUNCTION__);
            return;
        }
        global_pSdio->card->dev = (struct device *) dev;
        pNotifier->add(global_pSdio->card);
    }
}

void
sdio_card_removed()
{
    card_is_inserted = 0;

                                                                                                                                                                                                /*--- mmoser 4/4/2006 ---*/
    if (global_pSdio == NULL) {
        printk("%s: global_pSdio == NULL !!!!\n", __FUNCTION__);
        return;
    }

    if (pNotifier != NULL && pNotifier->remove != NULL) {
        DBGPRINT(DBG_HWIF,
                 (KERN_DEBUG
                  "%s called :pNotifier->remove=%p global_pSdio=%p card_is_removed=%d\n",
                  __FUNCTION__, pNotifier->remove, global_pSdio,
                  card_is_removed));
                                                                                                                                                                                                                /*--- mmoser 3/31/2006 ---*/
        if (card_is_removed == 0 && global_pSdio) {
            pNotifier->remove(global_pSdio->card);
        }
    }
}

void *
sdio_register(int type, void *ops, int memory)
{

    void *ret = NULL;

    switch (type) {

    case MMC_TYPE_HOST:
        DBGPRINT(DBG_ERROR,
                 (KERN_DEBUG
                  "sdio_register: call with MMC_TYPE_HOST not supported!!!!!\n"));
        break;

    case MMC_REG_TYPE_USER:

                                                                                                                                                                        /*--- mmoser 3/30/2006 ---*/
        pNotifier = (mmc_notifier_t) ops;

        if (sdio_CreateIface(&global_pSdio, NULL, NULL, NULL) ==
            SDIO_API_STATUS_SUCCESS) {
            //  pSdio is the pointer to the SDIO API interface structure - pass it
            //  to the wlan module. Will be needed for further calls to the SDIO
            //  interface.
            ret = (void *) global_pSdio;
        }
                                                                                                                                                                                                                        /*--- mmoser 3/31/2006 ---*/
        mrvlsdio_register_callbacks(sdio_card_added, sdio_card_removed);
        ret = (void *) 1;
        break;

    default:
        break;
    }

    return ret;
}

int
sdio_unregister(int type, void *ops)
{

    if (global_pSdio == NULL) {
        DBGPRINT(DBG_ERROR,
                 (KERN_DEBUG
                  "sdio_unregister: global_pSdio == NULL !!!!!\n"));

                                                                                                                                                                        /*--- mmoser 4/4/2006 ---*/
        pNotifier = NULL;
        mrvlsdio_register_callbacks(NULL, NULL);
        return 1;
    }

    switch (type) {
    case MMC_TYPE_HOST:
        break;
    case MMC_REG_TYPE_USER:
        if (card_is_inserted == 1 &&
            pNotifier != NULL && pNotifier->remove != NULL) {
            pNotifier->remove(global_pSdio->card);
        }

        if (global_pSdio->card)
            kfree(global_pSdio->card);

        if (global_pSdio->tmpl)
            kfree(global_pSdio->tmpl);

                                                                                                                                                                /*--- mmoser 3/31/2006 ---*/
        pNotifier = NULL;
        mrvlsdio_register_callbacks(NULL, NULL);

        sdio_DestroyIface(&global_pSdio);

                                                                                                                                                                                                /*--- mmoser 4/4/2006 ---*/
        global_pSdio = NULL;
        break;

    default:
        break;
    }

    return 0;
}

int
sdio_free_irq(mmc_card_t card, void *dev_id)
{
    int ret;

//      _ENTER();

                                                                                                                                                                                /*--- mmoser 3/31/2006 ---*/
    card_is_removed = 1;

    if (!card) {
        ret = -ENODEV;
        goto done;
    }

    /* Disable function IOs on the card */
    sdio_write_ioreg(card, FN0, IO_ENABLE_REG, 0x0);

    /* Disable function interrupts on the function */
    sdio_write_ioreg(card, FN0, INT_ENABLE_REG, 0x0);

//      printk("%s: card=%p  card->ctrlr=%p line=%d\n",__FUNCTION__,card,card->ctrlr,__LINE__);                         
    sdio_Register_ISR_Handler(card->ctrlr, NULL, NULL);

    /* Release IRQ from OS */

    return 0;

  done:

    return 0;
}

int
sdio_get_vendor_id(mmc_card_t card)
{
    card->info.cisptr[0] = card_GetCIS_Pointer(card->ctrlr, 0);
    card->info.cisptr[1] = card_GetCIS_Pointer(card->ctrlr, 1);

    return (int) 0x02df;
}

int
start_bus_clock(mmc_controller_t ctrlr)
{
    DBGPRINT(DBG_HWIF, (KERN_DEBUG "%s", __FUNCTION__));
    sdio_SetClock(ctrlr, 1);
    return 0;
}

int
stop_bus_clock_2(mmc_controller_t ctrlr)
{
    DBGPRINT(DBG_HWIF, (KERN_DEBUG "%s", __FUNCTION__));
    sdio_SetClock(ctrlr, 0);
    return 0;
}

int
sdio_check_idle_state(mmc_controller_t ctrlr)
{
    return 0;
}

void
sdio_print_imask(mmc_controller_t ctrlr)
{
}

void
sdio_clear_imask(mmc_controller_t ctrlr)
{
}

void
sdio_unmask(void)
{
    sdio_EnableInterrupt(global_pSdio);
}

EXPORT_SYMBOL(sdio_CreateIface);
EXPORT_SYMBOL(sdio_DestroyIface);
EXPORT_SYMBOL(sdio_Read_Byte);
EXPORT_SYMBOL(sdio_Write_Byte);
EXPORT_SYMBOL(sdio_Read_Block);
EXPORT_SYMBOL(sdio_Write_Block);

                                                                                                                                                                        /*--- mmoser 3/20/2006 ---*/
EXPORT_SYMBOL(sdio_Read_BlockEx);
EXPORT_SYMBOL(sdio_Write_BlockEx);

                                                                                                                                                                        /*--- mmoser 3/22/2006 ---*/
EXPORT_SYMBOL(sdio_clear_imask);
EXPORT_SYMBOL(sdio_check_idle_state);
EXPORT_SYMBOL(sdio_print_imask);
EXPORT_SYMBOL(stop_bus_clock_2);

                                                                                                                                                                        /*--- mmoser 3/29/2006 ---*/
EXPORT_SYMBOL(sdio_free_irq);
EXPORT_SYMBOL(sdio_register);
EXPORT_SYMBOL(sdio_get_vendor_id);
EXPORT_SYMBOL(sdio_unmask);
EXPORT_SYMBOL(sdio_unregister);
EXPORT_SYMBOL(start_bus_clock);
                                                                                                                                                                        /*--- mmoser 2/20/2007 ---*/
EXPORT_SYMBOL(sdio_SetGPO);

EXPORT_SYMBOL(sdio_SetClock);
EXPORT_SYMBOL(sdio_SetBusWidth);
EXPORT_SYMBOL(sdio_SetClockSpeed);

EXPORT_SYMBOL(sdio_isCardPresent);
EXPORT_SYMBOL(sdio_lockInterface);
EXPORT_SYMBOL(sdio_unlockInterface);
EXPORT_SYMBOL(sdio_SetLed);
EXPORT_SYMBOL(sdio_PsState);
EXPORT_SYMBOL(sdio_EnableInterrupt);
EXPORT_SYMBOL(sdio_DisableInterrupt);
EXPORT_SYMBOL(sdio_Register_ISR_Handler);
EXPORT_SYMBOL(card_isCardReady);
EXPORT_SYMBOL(card_disable_card_int);
EXPORT_SYMBOL(card_enable_card_int);
EXPORT_SYMBOL(card_Helper_Download);
EXPORT_SYMBOL(card_Firmware_Download);
EXPORT_SYMBOL(fw_CardToHost);
EXPORT_SYMBOL(fw_HostToCard);
EXPORT_SYMBOL(ps_sleep_confirmed);

                                                                                                                                                                        /*--- mmoser 3/31/2006 ---*/
EXPORT_SYMBOL(sdio_card_added);
EXPORT_SYMBOL(sdio_card_removed);

#endif
