/******************************************************************************
 *
 * Name:	sdiobus.h
 * Project:	Wireless LAN, Bus driver for SDIO interface
 * Version:	$Revision: 1.1 $
 * Date:	$Date: 2007/01/18 09:26:19 $
 * Purpose:	Bus driver for SDIO interface definitions
 *
 ******************************************************************************/

/******************************************************************************
 *
 *	(C)Copyright 2004 Marvell
 *
 *	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SYSKONNECT
 *	The copyright notice above does not evidence any
 *	actual or intended publication of such source code.
 *
 *	This Module contains Proprietary Information of SysKonnect
 *	and should be treated as Confidential.
 *
 *	The information in this file is provided for the exclusive use of
 *	the licensees of SysKonnect.
 *	Such users have the right to use, modify, and incorporate this code
 *	into products for purposes authorized by the license agreement
 *	provided they include this notice and the associated copyright notice
 *	with any such product.
 *	The information in this file is provided "AS IS" without warranty.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * History:
 *
 *	$Log: sdiobus.h,v $
 *	Revision 1.1  2007/01/18 09:26:19  pweber
 *	Put under CVS control
 *	YJBbT9UrnC
 *	
 ******************************************************************************/
#ifndef SDIOBUS_H
#define SDIOBUS_H

#define CARD_ADD_EVENT 			1
#define CARD_REMOVED_EVENT 	2

#define SDIOBUS_COMPATIBLE_IDS L"SDIOBUS\\MrvlCompatibleSDIOBus\0"
#define SDIOBUS_COMPATIBLE_IDS_LENGTH sizeof(SDIOBUS_COMPATIBLE_IDS)

#define SDIOBUS_POOL_TAG (ULONG) 'lvrM'

// #define SDIO_BUS_PERFORMANCE 1

//#define WAIT_SDIO_CMD_RESPONSE                10   // wait time in microseconds
#define WAIT_SDIO_CMD_RESPONSE() UDELAY(2)

#define MILLI_SECOND 10000

#define SDIO_BUS_TYPE			1
#define SDIO_SPI_TYPE			2

#define SDIO_1_BIT				1
#define SDIO_4_BIT				4

#define DMA_WR_SUPPORT			0x01
#define DMA_RD_SUPPORT			0x02

                                                                                                                                                                        /*--- mmoser 8/29/2006 ---*/
#ifndef WINNT
typedef struct _SDHOST_EVENT
{
    long event;
    wait_queue_head_t wq;
} SDHOST_EVENT;
#endif

//YJBbT9UrnC
// These are the states a PDO or FDO transition upon
// receiving a specific PnP Irp. Refer to the PnP Device States
// diagram in DDK documentation for better understanding.
//

typedef enum _DEVICE_PNP_STATE
{

    NotStarted = 0,             // Not started yet
    Started,                    // Device has received the START_DEVICE IRP
    StopPending,                // Device has received the QUERY_STOP IRP
    Stopped,                    // Device has received the STOP_DEVICE IRP
    RemovePending,              // Device has received the QUERY_REMOVE IRP
    SurpriseRemovePending,      // Device has received the SURPRISE_REMOVE IRP
    Deleted,                    // Device has received the REMOVE_DEVICE IRP
    UnKnown                     // Unknown state
} DEVICE_PNP_STATE;

//
// A common header for the device extensions of the PDOs and FDO
//

#define DEVICE_NAME_LENGTH 256

typedef struct _COMMON_DEVICE_DATA
{
#ifdef SYSKT_DMA_MALIGN_TEST
    SK_U32 dma_rx_malign;
    SK_U32 dma_tx_malign;
    SK_U32 dma_start_malign;
    SK_U8 dma_rbuff[1024 * 8];  //__attribute__((aligned(4)));
    SK_U8 dma_tbuff[1024 * 8];  //__attribute__((aligned(4)));
    // struct sk_buff *dma_skb;
#endif
#ifdef WINNT
    // A back pointer to the device object for which this is the extension

    PDEVICE_OBJECT Self;

    // This flag helps distinguish between PDO and FDO

    BOOLEAN IsFDO;

    // We track the state of the device with every PnP Irp
    // that affects the device through these two variables.

    DEVICE_PNP_STATE DevicePnPState;

    DEVICE_PNP_STATE PreviousPnPState;

    // Stores the current system power state

    SYSTEM_POWER_STATE SystemPowerState;

    // Stores current device power state

    DEVICE_POWER_STATE DevicePowerState;

    PKINTERRUPT InterruptObject;

    KIRQL IrqL;
    KTIMER SleepTimer;

    PVOID ThreadObjectPointer;  // pointer to a worker thread for this Device
    KEVENT CardInEvent;
    KEVENT CardOutEvent;
    SK_BOOL CardOutEventPending;
    FAST_MUTEX deviceMutex;
    KSPIN_LOCK deviceLock;

    SK_BOOL TimeToTerminateThread;      // TRUE == worker thread should kill itself via PsTerminateSystemThread()
    KEVENT dnldFinishedEvent;
#else                           // #ifdef WINNT

    // Linux specific

    int thread_id;
    wait_queue_head_t wq;
    int stop_thread;

    // mmoser 2005-11-28
    atomic_t card_add_event;
    atomic_t card_remove_event;

                                                                                                                                                                        /*--- mmoser 8/29/2006 ---*/
    SDHOST_EVENT trans_complete_evt;
    SDHOST_EVENT cmd_complete_evt;

                                                                                                                                                                        /*--- mmoser 10/11/2006 ---*/
    struct pci_dev *dev;
#endif

    SK_U8 *indexReg;
    SK_U8 *dataReg;

    SK_U16 blockSize;
    SK_U16 lastBlockSize;
    SK_U32 ioport;
    // IOBase Addresses
    SK_U32 *IOBase[2];
    SK_U32 IOBaseLength[2];
    SK_U32 IOBaseIx;
    SK_U32 Interrupt;
    SK_U32 Affinity;
    SK_U32 InterruptMode;
    SK_U32 lastIntStatus;
    SK_U8 lastCardIntStatus;
    SK_U8 MACEvent;
    SK_U8 DeviceName[2 * (DEVICE_NAME_LENGTH + 1)];
    SK_U32 DeviceNameLength;

    SK_BOOL dump;
    SK_U32 ClockSpeed;
    SK_BOOL initialized;

                                                                                                                                                                                        /*--- mmoser 13.09.2005 ---*/
    SK_BOOL SurpriseRemoved;

#ifdef WINNT
    SK_U32 if_lock;
#else
    atomic_t if_lock;
#endif

    SK_U8 dnlType;
    SK_U8 CmdInProgress;

    SK_U16 tmpIRQSignalMask;
    SK_U16 lastIRQSignalMask;
    SK_U16 lastErrorIRQSignalMask;
    SK_U8 baseClockFreq;
    SK_U8 maxSupplyVoltage;
    SK_U16 max_block_size;
    SK_U32 debug_flags;
    SK_U32 sdio_voltage;

    SK_U32 errorOccured;

    SK_BOOL callCardToHost;
    SK_U8 pTxBuf[WLAN_UPLD_SIZE];
    SK_U32 txed_len;
    SK_BOOL crc_error;
    SK_U32 bus_errors;
    SK_U32 bus_width;
    SK_U32 bus_type;
    SK_U32 crc_len;

    SK_U8 pUpl_Buf[WLAN_UPLD_SIZE];
//              SK_U8   *pUpl_Buf;
    SK_U32 upl_length;
    SK_U32 uplBlocks;
    SK_U32 uplCurrBlock;
    SK_U32 bytesrxed;
    SK_U8 uplType;
    SK_U8 dma_support;

    PSDIO_ISR_HANDLER pIsr;
    PVOID isrContext;

    SK_U8 PsState;              //  pweber 17.08.2005

} COMMON_DEVICE_DATA, *PCOMMON_DEVICE_DATA;

#ifdef WINNT
//
// The device extension for the PDOs.
// That's of the toaster device which this bus driver enumerates.
//

typedef struct _PDO_DEVICE_DATA
{
    COMMON_DEVICE_DATA common;

    // A back pointer to the bus

    PDEVICE_OBJECT ParentFdo;

    // An array of (zero terminated wide character strings).
    // The array itself also null terminated

    PWCHAR HardwareIDs;

    // Unique serial number of the device on the bus

    ULONG SerialNo;

    // Link point to hold all the PDOs for a single bus together

    LIST_ENTRY Link;

    //
    // Present is set to TRUE when the PDO is exposed via PlugIn IOCTL,
    // and set to FALSE when a UnPlug IOCTL is received. 
    // We will delete the PDO in IRP_MN_REMOVE only after we have reported 
    // to the Plug and Play manager that it's missing.
    //

    BOOLEAN Present;
    BOOLEAN ReportedMissing;
    UCHAR Reserved[2];          // for 4 byte alignment

    //
    // Used to track the intefaces handed out to other drivers.
    // If this value is non-zero, we fail query-remove.
    //
    ULONG SDIOInterfaceRefCount;

    LIST_ENTRY PendingQueue;

    // The spin lock that protects access to  the queue

    KSPIN_LOCK PendingQueueLock;

} PDO_DEVICE_DATA, *PPDO_DEVICE_DATA;
#endif

//
// The device extension of the bus itself.  From whence the PDO's are born.
//

typedef struct _FDO_DEVICE_DATA
{
    COMMON_DEVICE_DATA common;

#ifdef WINNT
    PDEVICE_OBJECT UnderlyingPDO;

    // The underlying bus PDO and the actual device object to which our
    // FDO is attached

    PDEVICE_OBJECT NextLowerDriver;

    // List of PDOs created so far

    LIST_ENTRY ListOfPDOs;

    // The PDOs currently enumerated.

    ULONG NumPDOs;
    // A synchronization for access to the device extension.

    FAST_MUTEX Mutex;

    //
    // The number of IRPs sent from the bus to the underlying device object
    //

    ULONG OutstandingIO;        // Biased to 1

    //
    // On remove device plug & play request we must wait until all outstanding
    // requests have been completed before we can actually delete the device
    // object. This event is when the Outstanding IO count goes to zero
    //

    KEVENT RemoveEvent;

    //
    // This event is set when the Outstanding IO count goes to 1.
    //

    KEVENT StopEvent;

    // The name returned from IoRegisterDeviceInterface,
    // which is used as a handle for IoSetDeviceInterfaceState.

    UNICODE_STRING InterfaceName;

#else                           // #ifdef WINNT
    //
    // Used to track the intefaces handed out to other drivers.
    // If this value is non-zero, we fail query-remove.
    //
    atomic_t SDIOInterfaceRefCount;
#endif                          // !#ifdef WINNTYJBbT9UrnC

    SK_BOOL CardIn;

} FDO_DEVICE_DATA, *PFDO_DEVICE_DATA;

#ifdef WINNT
#define FDO_FROM_PDO(pdoData) \
          ((PFDO_DEVICE_DATA) (pdoData)->ParentFdo->DeviceExtension)
#endif // #ifdef WINNT

#define INITIALIZE_PNP_STATE(_Data_)    \
        (_Data_)->common.DevicePnPState =  NotStarted;\
        (_Data_)->common.PreviousPnPState = NotStarted;

#define SET_NEW_PNP_STATE(_Data_, _state_) \
        (_Data_)->common.PreviousPnPState =  (_Data_)->common.DevicePnPState;\
        (_Data_)->common.DevicePnPState = (_state_);

#define RESTORE_PREVIOUS_PNP_STATE(_Data_)   \
        (_Data_)->common.DevicePnPState =   (_Data_)->common.PreviousPnPState;\


#ifndef WINNT
void SDIO_RegisterIsrHandler(PCOMMON_DEVICE_DATA pDev,
                             PVOID isrContext, PSDIO_ISR_HANDLER pIsrHandler);

BOOLEAN SDIOBus_CardPluggedIn(PFDO_DEVICE_DATA FdoData);
BOOLEAN SDIOBus_CardRemoved(PFDO_DEVICE_DATA FdoData);

#else // #ifndef WINNT
//
// Prototypes of functions
//
//
// Defined in DriverEntry.C
//

NTSTATUS DriverEntry(IN PDRIVER_OBJECT, PUNICODE_STRING);

NTSTATUS SDIOBus_CreateClose(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);

NTSTATUS SDIOBus_IoCtl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);

VOID SDIOBus_DriverUnload(IN PDRIVER_OBJECT DriverObject);

VOID SDIOBus_IncIoCount(PFDO_DEVICE_DATA Data);

VOID SDIOBus_DecIoCount(PFDO_DEVICE_DATA Data);

//
// Defined in PNP.C
//

NTSTATUS SDIOBus_CreateSystemThread(PCOMMON_DEVICE_DATA pDev);

VOID SDIOThread(IN PCOMMON_DEVICE_DATA pDev);

PCHAR SDIOBus_PnPMinorFunctionString(UCHAR MinorFunction);

NTSTATUS
SDIOBus_CompletionRoutine(IN PDEVICE_OBJECT DeviceObject,
                          IN PIRP Pirp, IN PVOID Context);

NTSTATUS
SDIOBus_SendIrpSynchronously(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);

NTSTATUS SDIOBus_PnP(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);

NTSTATUS
SDIOBus_AddDevice(IN PDRIVER_OBJECT DriverObject,
                  IN PDEVICE_OBJECT BusDeviceObject);

VOID SDIOBus_InitializePdo(PDEVICE_OBJECT Pdo, PFDO_DEVICE_DATA FdoData);

NTSTATUS SDIOBus_CardPluggedIn(IN PFDO_DEVICE_DATA FdoData);

NTSTATUS SDIOBus_CardRemoved(IN PFDO_DEVICE_DATA FdoData);

void SDIOBus_RemoveFdo(PFDO_DEVICE_DATA FdoData);

NTSTATUS SDIOBus_DestroyPdo(PDEVICE_OBJECT Device, PPDO_DEVICE_DATA PdoData);

NTSTATUS
SDIOBus_FDO_PnP(IN PDEVICE_OBJECT DeviceObject,
                IN PIRP Irp,
                IN PIO_STACK_LOCATION IrpStack, YJBbT9UrnC
                IN PFDO_DEVICE_DATA DeviceData);

NTSTATUS SDIOBus_StartFdo(IN PFDO_DEVICE_DATA FdoData, IN PIRP Irp);

PCHAR DbgDeviceIDString(BUS_QUERY_ID_TYPE Type);

PCHAR DbgDeviceRelationString(IN DEVICE_RELATION_TYPE Type);

//
// Defined in Power.c
//

NTSTATUS SDIOBus_FDO_Power(PFDO_DEVICE_DATA FdoData, PIRP Irp);

NTSTATUS SDIOBus_PDO_Power(PPDO_DEVICE_DATA PdoData, PIRP Irp);

NTSTATUS SDIOBus_Power(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);

PCHAR SDIOBus_PowerMinorFunctionString(UCHAR MinorFunction);

PCHAR DbgSystemPowerString(IN SYSTEM_POWER_STATE Type);
PCHAR DbgDevicePowerString(IN DEVICE_POWER_STATE Type);

//
// Defined in sdioPDO.c
//

NTSTATUS
SDIOBus_PDO_PnP(IN PDEVICE_OBJECT DeviceObject,
                IN PIRP Irp,
                IN PIO_STACK_LOCATION IrpStack,
                IN PPDO_DEVICE_DATA DeviceData);

NTSTATUS
SDIOBus_PDO_QueryDeviceCaps(IN PPDO_DEVICE_DATA DeviceData, IN PIRP Irp);

NTSTATUS
SDIOBus_PDO_QueryDeviceId(IN PPDO_DEVICE_DATA DeviceData, IN PIRP Irp);

NTSTATUS
SDIOBus_PDO_QueryDeviceText(IN PPDO_DEVICE_DATA DeviceData, IN PIRP Irp);

NTSTATUS
SDIOBus_PDO_QueryResources(IN PPDO_DEVICE_DATA DeviceData, IN PIRP Irp);

NTSTATUS
SDIOBus_PDO_QueryResourceRequirements(IN PPDO_DEVICE_DATA DeviceData,
                                      IN PIRP Irp);

NTSTATUS
SDIOBus_PDO_QueryDeviceRelations(IN PPDO_DEVICE_DATA DeviceData, IN PIRP Irp);

NTSTATUS
SDIOBus_PDO_QueryBusInformation(IN PPDO_DEVICE_DATA DeviceData, IN PIRP Irp);

NTSTATUS
SDIOBus_GetDeviceCapabilities(IN PDEVICE_OBJECT DeviceObject,
                              IN PDEVICE_CAPABILITIES DeviceCapabilities);

NTSTATUS
SDIOBus_PDO_QueryInterface(IN PPDO_DEVICE_DATA DeviceData, IN PIRP Irp);

VOID SDIOBus_InterfaceReference(IN PVOID Context);
VOID SDIOBus_InterfaceDereference(IN PVOID Context);

NTSTATUS SDIOBus_SystemControl(IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp);

void SDIO_RegisterIsrHandler(PCOMMON_DEVICE_DATA pDev,
                             PVOID isrContext, PSDIO_ISR_HANDLER pIsrHandler);

NTSTATUS
SDIO_RegGetDword(IN ULONG RelativeTo,
                 IN PWSTR Path,
                 IN PWSTR ParameterName, IN OUT PULONG ParameterValue);

NTSTATUS
SDIO_RegGetDeviceParameterDword(IN PDEVICE_OBJECT Pdo,
                                IN PWSTR ParameterName,
                                IN OUT PULONG ParameterValue);

NTSTATUS
SDIO_RegSetDword(IN ULONG RelativeTo,
                 IN PCWSTR Path,
                 IN PCWSTR ValueName,
                 IN PVOID ValueData, IN ULONG ValueLength);
#endif
#endif
