/******************************************************************************
 *
 * Name:	sdiobus.c
 * Project:	Wireless LAN, Bus driver for SDIO interface
 * Version:	$Revision: 1.1 $
 * Date:	$Date: 2007/01/18 09:21:35 $
 * Purpose:	This module handles the interrupts.
 *
 ******************************************************************************/

/******************************************************************************
 *
 *	(C)Copyright 2004 Marvell
 *
 *	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF SYSKONNECT
 *	The copyright notice above does not evidence any
 *	actual or intended publication of such source code.
 *
 *	This Module contains Proprietary Information of SysKonnect
 *	and should be treated as Confidential.
 *
 *	The information in this file is provided for the exclusive use of
 *	the licensees of SysKonnect.
 *	Such users have the right to use, modify, and incorporate this code
 *	into products for purposes authorized by the license agreement
 *	provided they include this notice and the associated copyright notice
 *	with any such product.
 *	The information in this file is provided "AS IS" without warranty.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * History:
 *
 *	$Log: skisr.c,v $
 *	Revision 1.1  2007/01/18 09:21:35  pweber
 *	Put under CVS control
 *	
 *
 ******************************************************************************/

#include "h/skdrv1st.h"
#include "h/skdrv2nd.h"

#ifdef WINNT
VOID
SDIOBus_Dpc(IN PKDPC Dpc,
            IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp, IN PVOID Context)
{
    SK_U8 is, rsr_value, ucValue = 0;
    SK_U32 irq_status, status, j, ulVal, ulDataLineMask, loops;
    SK_U16 usValue;
    PCOMMON_DEVICE_DATA commonData;
    PFDO_DEVICE_DATA FdoData;

    DBGPRINT(DBG_IRQ, ("SDIOBUS_Dpc\n"));

    if (DeviceObject == NULL) {
        DBGPRINT(DBG_ERROR, ("SDIOBUS_Dpc : DeviceObject == NULL !\n"));
        return;
    }

    FdoData = (PFDO_DEVICE_DATA) DeviceObject->DeviceExtension;

    if (FdoData == NULL) {
        DBGPRINT(DBG_ERROR, ("SDIOBUS_Dpc : FdoData == NULL !\n"));
        return;
    }

    irq_status = FdoData->common.lastIntStatus;

    if ((irq_status & STDHOST_NORMAL_IRQ_CARD_OUT) ==
        STDHOST_NORMAL_IRQ_CARD_OUT) {
        // clear irq bit
        MEM_WRITE_USHORT((PCOMMON_DEVICE_DATA) FdoData, SK_SLOT_0,
                         STDHOST_NORMAL_IRQ_STATUS,
                         STDHOST_NORMAL_IRQ_CARD_OUT);

                                                                                                                                                                        /*--- mmoser 14.09.2005 ---*/
        if (FdoData->common.initialized) {
            DmBGPRINT(DBG_IRQ | DBG_W528D,
                      ("SDIOBUS_Dpc : NORMAL_IRQ_CARD_OUT\n"));

                                                                                                                                                                                /*--- mmoser 10.08.2005 ---*/
            DBGPRINT(DBG_IRQ, ("CARD REMOVED\n"));
            KeSetEvent(&FdoData->common.CardOutEvent, 0, FALSE);
        }
        return;
    }
    if ((irq_status & STDHOST_NORMAL_IRQ_CARD_IN) ==
        STDHOST_NORMAL_IRQ_CARD_IN) {
        // clear irq bit
        MEM_WRITE_USHORT((PCOMMON_DEVICE_DATA) FdoData, SK_SLOT_0,
                         STDHOST_NORMAL_IRQ_STATUS,
                         STDHOST_NORMAL_IRQ_CARD_IN);

        DBGPRINT(DBG_IRQ, ("SDIOBUS_Dpc : NORMAL_IRQ_CARD_IN\n"));
        DBGPRINT(DBG_IRQ, ("CARD PLUGGED IN\n"));
        KeSetEvent(&FdoData->common.CardInEvent, 0, FALSE);
        return;
    }
    if ((irq_status & STDHOST_NORMAL_IRQ_CARD) == STDHOST_NORMAL_IRQ_CARD) {
        PCOMMON_DEVICE_DATA pDev = (PCOMMON_DEVICE_DATA) FdoData;

        status = (irq_status << 16);

        if (FdoData->common.pIsr != NULL) {
            DBGPRINT(DBG_IRQ, ("Call registered interrupt handler.\n"));
            (*FdoData->common.pIsr) (FdoData->common.isrContext, status);
        } else {
            DBGPRINT(DBG_IRQ,
                     ("No interrupt handler registered. (fdo=%p)\n",
                      FdoData));
            SDHost_EnableInterrupt((PCOMMON_DEVICE_DATA) FdoData,
                                   STDHOST_NORMAL_IRQ_CARD_ALL_ENA);
            return;
        }

        MEM_READ_ULONG(&FdoData->common, SK_SLOT_0, STDHOST_PRESENT_STATE,
                       &ulVal);
        if (FdoData->common.bus_type == SDIO_SPI_TYPE ||
            ((FdoData->common.bus_type == SDIO_BUS_TYPE) &&
             (FdoData->common.bus_width == SDIO_1_BIT))) {
            if ((ulVal & STDHOST_STATE_DAT1_LINE_LVL) == 0) {
                DBGPRINT(DBG_IRQ,
                         ("SDHost_Dpc() : Interrupt Line is active !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"));
            }
        }
    } else {
        MEM_WRITE_USHORT((PCOMMON_DEVICE_DATA) FdoData, SK_SLOT_0,
                         STDHOST_NORMAL_IRQ_STATUS, 0xFFFF);
        DBGPRINT(DBG_ERROR,
                 ("SDIOBus_Dpc(): Unhandled IRQ : 0x%8.08X.\n", irq_status));
    }

    SDHost_EnableInterrupt((PCOMMON_DEVICE_DATA) FdoData,
                           STDHOST_NORMAL_IRQ_CARD_ALL_ENA);
}

BOOLEAN
SDIOBus_Isr(IN PKINTERRUPT Interrupt, IN PVOID Context)
{
    SK_U32 tmp;
    SK_U8 ucValue, rsr_value;
    PDEVICE_OBJECT deviceObject = (PDEVICE_OBJECT) Context;
    PCOMMON_DEVICE_DATA commonData;

    if (deviceObject == NULL) {
        DBGPRINT(DBG_IRQ, ("SDIOBUS_Isr : deviceObject == NULL !\n"));
        return SK_TRUE;
    }

    commonData = (PCOMMON_DEVICE_DATA) deviceObject->DeviceExtension;

    if (commonData == NULL) {
        DBGPRINT(DBG_IRQ, ("SDIOBUS_Isr : commonData == NULL !\n"));
        return SK_TRUE;
    }
    // Clear ISR status
    MEM_READ_ULONG(commonData, SK_SLOT_0, STDHOST_NORMAL_IRQ_STATUS, &tmp);

    DBGPRINT(DBG_IRQ, ("I%8.08X (%s:0x%8.08X)\n", tmp,
                       commonData->IsFDO ? "FDO" : "PDO",
                       commonData->lastIRQSignalMask));

    // Check if interrupt came from our device
    if (((tmp & STDHOST_NORMAL_IRQ_ERROR) != 0 ||
         (tmp & commonData->lastIRQSignalMask) != 0) && tmp != 0xFFFFFFFF) {
        PCOMMON_DEVICE_DATA pDev = (PCOMMON_DEVICE_DATA) commonData;
        commonData->lastIntStatus = tmp;
        DBGPRINT(DBG_IRQ,
                 ("SDIOBUS_Isr (0x%4.04X)\n", commonData->lastIntStatus));

        SDHost_DisableInterrupt(commonData, STDHOST_NORMAL_IRQ_ALL_SIG_ENA);
        IoRequestDpc(deviceObject, deviceObject->CurrentIrp, NULL);
        return SK_TRUE;
    } else {
        return SK_FALSE;
    }
}
#else // #ifdef WINNT

static void (*gpio_int_callback) (void *) = NULL;
static void *gpio_int_callback_arg = NULL;

int
request_gpio_irq_callback(void (*callback) (void *), void *arg)
{
    if (!gpio_int_callback && !gpio_int_callback_arg) {
        gpio_int_callback = callback;
        gpio_int_callback_arg = arg;
        return 0;
    }
    return -1;
}

EXPORT_SYMBOL(request_gpio_irq_callback);

int
release_gpio_irq_callback(void (*callback) (void *), void *arg)
{
    if ((callback == gpio_int_callback)
        && (arg == gpio_int_callback_arg)) {
        gpio_int_callback = NULL;
        gpio_int_callback_arg = NULL;
        return 0;
    }
    return -1;
}

EXPORT_SYMBOL(release_gpio_irq_callback);

DECLARE_TASKLET(SDIOBus_tasklet, SDIOBus_Dpc, 0);

void
SDIOBus_Dpc(unsigned long arg)
{
    SK_U32 irq_status, status, ulVal;
    SK_U16 errorIRQ_status = 0;
    PFDO_DEVICE_DATA FdoData;
    PCOMMON_DEVICE_DATA pDev;

    DBGPRINT(DBG_IRQ, (KERN_DEBUG "SDIOBUS_Dpc startet ....\n"));

    if (arg == 0) {
        DBGPRINT(DBG_ERROR, (KERN_DEBUG "SDIOBUS_Dpc : arg == NULL !\n"));
        return;
    }

    FdoData = (PFDO_DEVICE_DATA) arg;

    if (FdoData == NULL) {
        DBGPRINT(DBG_ERROR, (KERN_DEBUG "SDIOBUS_Dpc : FdoData == NULL !\n"));
        return;
    }
    pDev = (PCOMMON_DEVICE_DATA) & FdoData->common;
    irq_status = FdoData->common.lastIntStatus;

    if ((irq_status & STDHOST_NORMAL_IRQ_ERROR) != 0) {
        MEM_READ_USHORT((PCOMMON_DEVICE_DATA) & FdoData->common, SK_SLOT_0,
                        STDHOST_ERROR_IRQ_STATUS, &errorIRQ_status);
        if ((errorIRQ_status & STDHOST_ERROR_IRQ_GPI_0) != 0) {
            if ((FdoData->common.
                 lastErrorIRQSignalMask & STDHOST_ERROR_IRQ_GPI_0) != 0) {
                // clear irq bit
                MEM_WRITE_USHORT((PCOMMON_DEVICE_DATA) & FdoData->common,
                                 SK_SLOT_0, STDHOST_ERROR_IRQ_STATUS,
                                 STDHOST_ERROR_IRQ_GPI_0);
            }
        }
        if ((errorIRQ_status & STDHOST_ERROR_IRQ_GPI_1) != 0) {
            if ((FdoData->common.
                 lastErrorIRQSignalMask & STDHOST_ERROR_IRQ_GPI_1) != 0) {
                printk("GPI-1 interrupt fired (0x%4.04X)\n", errorIRQ_status);
                                /*--- mmoser 3/12/2007 ---*/
                // clear irq bit
                MEM_WRITE_USHORT((PCOMMON_DEVICE_DATA) & FdoData->common,
                                 SK_SLOT_0, STDHOST_ERROR_IRQ_STATUS,
                                 STDHOST_ERROR_IRQ_GPI_1);

                SDHost_SetClock(pDev, 1);
                if (gpio_int_callback && gpio_int_callback_arg)
                    gpio_int_callback(gpio_int_callback_arg);
            }
        }
    }

    if ((irq_status & STDHOST_NORMAL_IRQ_CARD_OUT) ==
        STDHOST_NORMAL_IRQ_CARD_OUT) {
        // clear irq bit
        MEM_WRITE_USHORT((PCOMMON_DEVICE_DATA) & FdoData->common, SK_SLOT_0,
                         STDHOST_NORMAL_IRQ_STATUS,
                         STDHOST_NORMAL_IRQ_CARD_OUT);

                /*--- mmoser 14.09.2005 ---*/
        if (FdoData->common.initialized) {
            DBGPRINT(DBG_IRQ | DBG_W528D,
                     (KERN_DEBUG "SDIOBUS_Dpc : NORMAL_IRQ_CARD_OUT\n"));

            /*--- mmoser 10.08.2005 ---*/
            DBGPRINT(DBG_IRQ, (KERN_DEBUG "CARD REMOVED\n"));
            SDIOBus_CardRemoved(FdoData);

                        /*--- mmoser 3/31/2006 ---*/
            FdoData->common.SurpriseRemoved = SK_TRUE;
        }
        SDHost_EnableInterrupt((PCOMMON_DEVICE_DATA) & FdoData->common,
                               STDHOST_NORMAL_IRQ_CARD_ALL_ENA);
        DBGPRINT(DBG_IRQ, (KERN_DEBUG "SDIOBUS_Dpc finished.\n"));
        return;
    }
    if ((irq_status & STDHOST_NORMAL_IRQ_CARD_IN) ==
        STDHOST_NORMAL_IRQ_CARD_IN) {
        // clear irq bit
        MEM_WRITE_USHORT((PCOMMON_DEVICE_DATA) & FdoData->common, SK_SLOT_0,
                         STDHOST_NORMAL_IRQ_STATUS,
                         STDHOST_NORMAL_IRQ_CARD_IN);

        DBGPRINT(DBG_IRQ, (KERN_DEBUG "SDIOBUS_Dpc : NORMAL_IRQ_CARD_IN\n"));
        DBGPRINT(DBG_IRQ, (KERN_DEBUG "CARD PLUGGED IN\n"));
        SDIOBus_CardPluggedIn(FdoData);
        SDHost_EnableInterrupt((PCOMMON_DEVICE_DATA) & FdoData->common,
                               STDHOST_NORMAL_IRQ_CARD_ALL_ENA);
        DBGPRINT(DBG_IRQ, (KERN_DEBUG "SDIOBUS_Dpc finished.\n"));
        return;
    }

    if ((irq_status & STDHOST_NORMAL_IRQ_CARD) == STDHOST_NORMAL_IRQ_CARD) {
        status = (irq_status << 16);

        if (FdoData->common.pIsr != NULL) {
            DBGPRINT(DBG_IRQ, (KERN_DEBUG "Call registered interrupt handler.\n"));                                                                                                                                                                                                             /*--- mmoser 3/20/2006 ---*/
            /*
             * reverse args order for compatibility of Marvell Linux drive
             */
            (*FdoData->common.pIsr) (status, FdoData->common.isrContext,
                                     NULL);
            return;
        } else {
            DBGPRINT(DBG_IRQ,
                     (KERN_DEBUG
                      "No interrupt handler registered. (fdo=%p)\n",
                      FdoData));
            SDHost_EnableInterrupt(&FdoData->common,
                                   STDHOST_NORMAL_IRQ_CARD_ALL_ENA);
            DBGPRINT(DBG_IRQ, (KERN_DEBUG "SDIOBUS_Dpc finished.\n"));
            return;
        }
        MEM_READ_ULONG(&FdoData->common, SK_SLOT_0, STDHOST_PRESENT_STATE,
                       &ulVal);
        if (FdoData->common.bus_type == SDIO_SPI_TYPE ||
            ((FdoData->common.bus_type == SDIO_BUS_TYPE) &&
             (FdoData->common.bus_width == SDIO_1_BIT))) {
            if ((ulVal & STDHOST_STATE_DAT1_LINE_LVL) == 0) {
                DBGPRINT(DBG_IRQ,
                         (KERN_DEBUG
                          "SDHost_Dpc() : Interrupt Line is active !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"));
            }
        }
    }

    if (irq_status & (STDHOST_NORMAL_IRQ_ERROR |
                      STDHOST_NORMAL_IRQ_TRANS_COMPLETE |
                      STDHOST_NORMAL_IRQ_CMD_COMPLETE)) {
        if (((irq_status & STDHOST_NORMAL_IRQ_TRANS_COMPLETE) &&
             (FdoData->common.
              tmpIRQSignalMask & STDHOST_NORMAL_IRQ_TRANS_COMPLETE)) ||
            ((irq_status & STDHOST_NORMAL_IRQ_CMD_COMPLETE) &&
             (FdoData->common.
              tmpIRQSignalMask & STDHOST_NORMAL_IRQ_CMD_COMPLETE)) ||
            ((errorIRQ_status & STDHOST_ERROR_IRQ_GPI_1) &&
             (FdoData->common.
              lastErrorIRQSignalMask & STDHOST_ERROR_IRQ_GPI_1))

            )
        {
            SDHost_EnableInterrupt((PCOMMON_DEVICE_DATA) & FdoData->common,
                                   STDHOST_NORMAL_IRQ_CARD_ALL_ENA);
        }
        return;
    } else {
        MEM_WRITE_USHORT((PCOMMON_DEVICE_DATA) & FdoData->common, SK_SLOT_0,
                         STDHOST_NORMAL_IRQ_STATUS, 0xFFFF);
        MEM_WRITE_USHORT((PCOMMON_DEVICE_DATA) & FdoData->common, SK_SLOT_0,
                         STDHOST_ERROR_IRQ_STATUS, 0xFFFF);
        DBGPRINT(DBG_ERROR,
                 (KERN_DEBUG "SDIOBus_Dpc(): Unhandled IRQ : 0x%8.08X.\n",
                  irq_status));
    }

    SDHost_EnableInterrupt((PCOMMON_DEVICE_DATA) & FdoData->common,
                           STDHOST_NORMAL_IRQ_CARD_ALL_ENA);
    DBGPRINT(DBG_IRQ, (KERN_DEBUG "SDIOBUS_Dpc finished.\n"));
}

irqreturn_t
SDIOBus_Isr(int irq, void *dev_id, struct pt_regs *regs)
{
    SK_U32 tmp;
    SK_U16 tmpsig;
    PFDO_DEVICE_DATA FdoData;
    PCOMMON_DEVICE_DATA pDev;

    if (dev_id == NULL) {
        DBGPRINT(DBG_IRQ, (KERN_DEBUG "SDIOBUS_Isr : dev_id == NULL !\n"));
        return IRQ_HANDLED;
    }

    FdoData = (PFDO_DEVICE_DATA) dev_id;

    if (FdoData == NULL) {
        DBGPRINT(DBG_IRQ, (KERN_DEBUG "SDIOBUS_Isr : FdoData == NULL !\n"));
        return IRQ_HANDLED;
    }

    pDev = (PCOMMON_DEVICE_DATA) & FdoData->common;
    MEM_READ_ULONG(&FdoData->common, SK_SLOT_0, STDHOST_NORMAL_IRQ_STATUS,
                   &tmp);
    MEM_READ_USHORT((PCOMMON_DEVICE_DATA) & FdoData->common, SK_SLOT_0,
                    STDHOST_ERROR_IRQ_SIGNAL_ENABLE, &tmpsig);

    // Check if interrupt came from our device
    if (((((tmp & STDHOST_NORMAL_IRQ_ERROR) != 0) &&
          ((tmpsig & (STDHOST_ERROR_IRQ_GPI_1 | STDHOST_ERROR_IRQ_GPI_0)) !=
           0)) || ((tmp & FdoData->common.lastIRQSignalMask) != 0)) &&
        tmp != 0xFFFFFFFF)
    {
                                                                                                                                                                /*--- mmoser 8/29/2006 ---*/
        if (tmp & STDHOST_NORMAL_IRQ_TRANS_COMPLETE) {
            if (FdoData->common.
                lastIRQSignalMask & STDHOST_NORMAL_IRQ_TRANS_COMPLETE) {
                // clear irq bit
                MEM_WRITE_USHORT((PCOMMON_DEVICE_DATA) & FdoData->common,
                                 SK_SLOT_0, STDHOST_NORMAL_IRQ_STATUS,
                                 STDHOST_NORMAL_IRQ_TRANS_COMPLETE);

                if (!test_and_set_bit
                    (0, &FdoData->common.trans_complete_evt.event)) {
                    wake_up(&FdoData->common.trans_complete_evt.wq);
                }
            }
        }

        if (tmp & STDHOST_NORMAL_IRQ_CMD_COMPLETE) {
            if (FdoData->common.
                lastIRQSignalMask & STDHOST_NORMAL_IRQ_CMD_COMPLETE) {
                // clear irq bit
                MEM_WRITE_USHORT((PCOMMON_DEVICE_DATA) & FdoData->common,
                                 SK_SLOT_0, STDHOST_NORMAL_IRQ_STATUS,
                                 STDHOST_NORMAL_IRQ_CMD_COMPLETE);

                if (!test_and_set_bit
                    (0, &FdoData->common.cmd_complete_evt.event)) {
                    wake_up(&FdoData->common.cmd_complete_evt.wq);
                }
            }
        }

        MEM_READ_USHORT((PCOMMON_DEVICE_DATA) & FdoData->common, SK_SLOT_0,
                        STDHOST_ERROR_IRQ_SIGNAL_ENABLE,
                        &FdoData->common.lastErrorIRQSignalMask);
        FdoData->common.lastIntStatus = tmp;
        DBGPRINT(DBG_IRQ,
                 (KERN_DEBUG "SDIOBUS_Isr (0x%4.04X)\n",
                  FdoData->common.lastIntStatus));

        SDHost_DisableInterrupt(&FdoData->common,
                                STDHOST_NORMAL_IRQ_ALL_SIG_ENA);

        SDIOBus_tasklet.data = (unsigned long) FdoData;
        tasklet_schedule(&SDIOBus_tasklet);

    } else {
        return IRQ_NONE;
    }
    return IRQ_HANDLED;

}
#endif // !#ifdef WINNT
